#
# Copyright 2018 www.privaz.io Valletech AB
#
# Simplified BSD License (see LICENSES/BSD-2-Clause.txt or https://opensource.org/licenses/BSD-2-Clause)
# SPDX-License-Identifier: BSD-2-Clause

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import ssl
import time
from os import environ

from ansible.module_utils.basic import AnsibleModule

IMAGE_STATES = [
    "INIT",
    "READY",
    "USED",
    "DISABLED",
    "LOCKED",
    "ERROR",
    "CLONE",
    "DELETE",
    "USED_PERS",
    "LOCKED_USED",
    "LOCKED_USED_PERS",
]
HAS_PYONE = True

try:
    from pyone import OneException
    from pyone.server import OneServer
except ImportError:
    OneException = Exception
    HAS_PYONE = False


# A helper function to mitigate https://github.com/OpenNebula/one/issues/6064.
# It allows for easily handling lists like "NIC" or "DISK" in the JSON-like template representation.
# There are either lists of dictionaries (length > 1) or just dictionaries.
def flatten(to_flatten, extract=False):
    """Flattens nested lists (with optional value extraction)."""

    def recurse(to_flatten):
        return sum(map(recurse, to_flatten), []) if isinstance(to_flatten, list) else [to_flatten]

    value = recurse(to_flatten)
    if extract and len(value) == 1:
        return value[0]
    return value


# A helper function to mitigate https://github.com/OpenNebula/one/issues/6064.
# It renders JSON-like template representation into OpenNebula's template syntax (string).
def render(to_render):
    """Converts dictionary to OpenNebula template."""

    def recurse(to_render):
        for key, value in sorted(to_render.items()):
            if value is None:
                continue
            if isinstance(value, dict):
                yield f"{key}=[{','.join(recurse(value))}]"
                continue
            if isinstance(value, list):
                for item in value:
                    yield f"{key}=[{','.join(recurse(item))}]"
                continue
            if isinstance(value, str):
                _value = value.replace("\\", "\\\\").replace('"', '\\"')
                yield f'{key}="{_value}"'
                continue
            yield f'{key}="{value}"'

    return "\n".join(recurse(to_render))


class OpenNebulaModule:
    """
    Base class for all OpenNebula Ansible Modules.
    This is basically a wrapper of the common arguments, the pyone client and
    some utility methods.
    """

    common_args = dict(
        api_url=dict(type="str", aliases=["api_endpoint"], default=environ.get("ONE_URL")),
        api_username=dict(type="str", default=environ.get("ONE_USERNAME")),
        api_password=dict(type="str", no_log=True, aliases=["api_token"], default=environ.get("ONE_PASSWORD")),
        validate_certs=dict(default=True, type="bool"),
        wait_timeout=dict(type="int", default=300),
    )

    def __init__(
        self, argument_spec, supports_check_mode=False, mutually_exclusive=None, required_one_of=None, required_if=None
    ):
        module_args = OpenNebulaModule.common_args.copy()
        module_args.update(argument_spec)

        self.module = AnsibleModule(
            argument_spec=module_args,
            supports_check_mode=supports_check_mode,
            mutually_exclusive=mutually_exclusive,
            required_one_of=required_one_of,
            required_if=required_if,
        )
        self.result = dict(changed=False, original_message="", message="")
        self.one = self.create_one_client()

        self.resolved_parameters = self.resolve_parameters()

    def create_one_client(self):
        """
        Creates an XMLPRC client to OpenNebula.

        Returns: the new xmlrpc client.

        """

        # context required for not validating SSL, old python versions won't validate anyway.
        if hasattr(ssl, "_create_unverified_context"):
            no_ssl_validation_context = ssl._create_unverified_context()
        else:
            no_ssl_validation_context = None

        # Check if the module can run
        if not HAS_PYONE:
            self.fail("pyone is required for this module")

        if self.module.params.get("api_url"):
            url = self.module.params.get("api_url")
        else:
            self.fail("Either api_url or the environment variable ONE_URL must be provided")

        if self.module.params.get("api_username"):
            username = self.module.params.get("api_username")
        else:
            self.fail("Either api_username or the environment variable ONE_USERNAME must be provided")

        if self.module.params.get("api_password"):
            password = self.module.params.get("api_password")
        else:
            self.fail("Either api_password or the environment variable ONE_PASSWORD must be provided")

        session = f"{username}:{password}"

        if not self.module.params.get("validate_certs") and "PYTHONHTTPSVERIFY" not in environ:
            return OneServer(url, session=session, context=no_ssl_validation_context)
        else:
            return OneServer(url, session)

    def close_one_client(self):
        """
        Close the pyone session.
        """
        self.one.server_close()

    def fail(self, msg):
        """
        Utility failure method, will ensure pyone is properly closed before failing.
        Args:
            msg: human readable failure reason.
        """
        if hasattr(self, "one"):
            self.close_one_client()
        self.module.fail_json(msg=msg)

    def exit(self):
        """
        Utility exit method, will ensure pyone is properly closed before exiting.

        """
        if hasattr(self, "one"):
            self.close_one_client()
        self.module.exit_json(**self.result)

    def resolve_parameters(self):
        """
        This method resolves parameters provided by a secondary ID to the primary ID.
        For example if cluster_name is present, cluster_id will be introduced by performing
        the required resolution

        Returns: a copy of the parameters that includes the resolved parameters.

        """

        resolved_params = dict(self.module.params)

        if "cluster_name" in self.module.params:
            clusters = self.one.clusterpool.info()
            for cluster in clusters.CLUSTER:
                if self.module.params.get("cluster_name") == cluster.NAME:
                    resolved_params["cluster_id"] = cluster.ID

        return resolved_params

    def is_parameter(self, name):
        """
        Utility method to check if a parameter was provided or is resolved
        Args:
            name: the parameter to check
        """
        if name in self.resolved_parameters:
            return self.get_parameter(name) is not None
        else:
            return False

    def get_parameter(self, name):
        """
        Utility method for accessing parameters that includes resolved ID
        parameters from provided Name parameters.
        """
        return self.resolved_parameters.get(name)

    def get_host_by_name(self, name):
        """
        Returns a host given its name.
        Args:
            name: the name of the host

        Returns: the host object or None if the host is absent.

        """
        hosts = self.one.hostpool.info()
        for h in hosts.HOST:
            if name == h.NAME:
                return h
        return None

    def get_cluster_by_name(self, name):
        """
        Returns a cluster given its name.
        Args:
            name: the name of the cluster

        Returns: the cluster object or None if the host is absent.
        """

        clusters = self.one.clusterpool.info()
        for c in clusters.CLUSTER:
            if name == c.NAME:
                return c
        return None

    def get_template_by_name(self, name):
        """
        Returns a template given its name.
        Args:
            name: the name of the template

        Returns: the template object or None if the host is absent.

        """
        templates = self.one.templatepool.info()
        for t in templates.TEMPLATE:
            if name == t.NAME:
                return t
        return None

    def cast_template(self, template):
        """
        OpenNebula handles all template elements as strings
        At some point there is a cast being performed on types provided by the user
        This function mimics that transformation so that required template updates are detected properly
        additionally an array will be converted to a comma separated list,
        which works for labels and hopefully for something more.

        Args:
            template: the template to transform

        Returns: the transformed template with data casts applied.
        """

        # TODO: check formally available data types in templates
        # TODO: some arrays might be converted to space separated

        for key in template:
            value = template[key]
            if isinstance(value, dict):
                self.cast_template(template[key])
            elif isinstance(value, list):
                template[key] = ", ".join(value)
            elif not isinstance(value, str):
                template[key] = str(value)

    def requires_template_update(self, current, desired):
        """
        This function will help decide if a template update is required or not
        If a desired key is missing from the current dictionary an update is required
        If the intersection of both dictionaries is not deep equal, an update is required
        Args:
            current: current template as a dictionary
            desired: desired template as a dictionary

        Returns: True if a template update is required
        """

        if not desired:
            return False

        self.cast_template(desired)
        intersection = dict()
        for dkey in desired.keys():
            if dkey in current.keys():
                intersection[dkey] = current[dkey]
            else:
                return True
        return desired != intersection

    def wait_for_state(
        self,
        element_name,
        state,
        state_name,
        target_states,
        invalid_states=None,
        transition_states=None,
        wait_timeout=None,
    ):
        """
        Args:
            element_name: the name of the object we are waiting for: HOST, VM, etc.
            state: lambda that returns the current state, will be queried until target state is reached
            state_name: lambda that returns the readable form of a given state
            target_states: states expected to be reached
            invalid_states: if any of this states is reached, fail
            transition_states: when used, these are the valid states during the transition.
            wait_timeout: timeout period in seconds. Defaults to the provided parameter.
        """

        if not wait_timeout:
            wait_timeout = self.module.params.get("wait_timeout")

        start_time = time.time()

        while (time.time() - start_time) < wait_timeout:
            current_state = state()

            if current_state in invalid_states:
                self.fail(f"invalid {element_name} state {state_name(current_state)}")

            if transition_states:
                if current_state not in transition_states:
                    self.fail(f"invalid {element_name} transition state {state_name(current_state)}")

            if current_state in target_states:
                return True

            time.sleep(self.one.server_retry_interval())

        self.fail(msg="Wait timeout has expired!")

    def run_module(self):
        """
        trigger the start of the execution of the module.
        Returns:

        """
        try:
            self.run(self.one, self.module, self.result)
        except OneException as e:
            self.fail(msg=f"OpenNebula Exception: {e}")

    def run(self, one, module, result):
        """
        to be implemented by subclass with the actual module actions.
        Args:
            one: the OpenNebula XMLRPC client
            module: the Ansible Module object
            result: the Ansible result
        """
        raise NotImplementedError("Method requires implementation")

    def get_image_list_id(self, image, element):
        """
        This is a helper function for get_image_info to iterate over a simple list of objects
        """
        list_of_id = []

        if element == "VMS":
            image_list = image.VMS
        if element == "CLONES":
            image_list = image.CLONES
        if element == "APP_CLONES":
            image_list = image.APP_CLONES

        for iter in image_list.ID:
            list_of_id.append(
                # These are optional so firstly check for presence
                getattr(iter, "ID", "Null"),
            )
        return list_of_id

    def get_image_snapshots_list(self, image):
        """
        This is a helper function for get_image_info to iterate over a dictionary
        """
        list_of_snapshots = []

        for iter in image.SNAPSHOTS.SNAPSHOT:
            list_of_snapshots.append(
                {
                    "date": iter["DATE"],
                    "parent": iter["PARENT"],
                    "size": iter["SIZE"],
                    # These are optional so firstly check for presence
                    "allow_orhans": getattr(image.SNAPSHOTS, "ALLOW_ORPHANS", "Null"),
                    "children": getattr(iter, "CHILDREN", "Null"),
                    "active": getattr(iter, "ACTIVE", "Null"),
                    "name": getattr(iter, "NAME", "Null"),
                }
            )
        return list_of_snapshots

    def get_image_info(self, image):
        """
        This method is used by one_image and one_image_info modules to retrieve
        information from XSD scheme of an image
        Returns: a copy of the parameters that includes the resolved parameters.
        """
        info = {
            "id": image.ID,
            "name": image.NAME,
            "state": IMAGE_STATES[image.STATE],
            "running_vms": image.RUNNING_VMS,
            "used": bool(image.RUNNING_VMS),
            "user_name": image.UNAME,
            "user_id": image.UID,
            "group_name": image.GNAME,
            "group_id": image.GID,
            "permissions": {
                "owner_u": image.PERMISSIONS.OWNER_U,
                "owner_m": image.PERMISSIONS.OWNER_M,
                "owner_a": image.PERMISSIONS.OWNER_A,
                "group_u": image.PERMISSIONS.GROUP_U,
                "group_m": image.PERMISSIONS.GROUP_M,
                "group_a": image.PERMISSIONS.GROUP_A,
                "other_u": image.PERMISSIONS.OTHER_U,
                "other_m": image.PERMISSIONS.OTHER_M,
                "other_a": image.PERMISSIONS.OTHER_A,
            },
            "type": image.TYPE,
            "disk_type": image.DISK_TYPE,
            "persistent": image.PERSISTENT,
            "regtime": image.REGTIME,
            "source": image.SOURCE,
            "path": image.PATH,
            "fstype": getattr(image, "FSTYPE", "Null"),
            "size": image.SIZE,
            "cloning_ops": image.CLONING_OPS,
            "cloning_id": image.CLONING_ID,
            "target_snapshot": image.TARGET_SNAPSHOT,
            "datastore_id": image.DATASTORE_ID,
            "datastore": image.DATASTORE,
            "vms": self.get_image_list_id(image, "VMS"),
            "clones": self.get_image_list_id(image, "CLONES"),
            "app_clones": self.get_image_list_id(image, "APP_CLONES"),
            "snapshots": self.get_image_snapshots_list(image),
            "template": image.TEMPLATE,
        }
        return info



#!/usr/bin/python
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

DOCUMENTATION = '''
---
module: azure_rm_ml_component_info
version_added: "3.16.0"
short_description: List ML Component resources
description:
    - List ML Component resources.
options:
    name:
        description:
            - Name of the Azure ML component.
        type: str
        required: false
    resource_group:
        description:
            - Name of resource group.
        required: true
        type: str
    ml_workspace:
        description:
            - Name of the Azure ML workspace.
        required: true
        type: str
    ml_registry:
        description:
            - If provided, the command will target the registry
              instead of a workspace.  Hence resource group and
              workspace won't be required. Must be provided if
              ml_workspace and resource_group are not provided.
        required: false
        type: str
    version:
        description:
            - Version of the data asset. Must be provided, if label
              is not provided. Mutually exclusive with label.
        required: false
        type: str
    label:
        description:
            - Label of the data asset. Must be provided, if version
              is not provided. Mutually exclusive with version.
        required: false
        type: str
    list_type:
        description:
            - Specify if you want active, archived or all datasets.
        type: str
        choices:
            - active
            - archived
            - all

extends_documentation_fragment:
    - azure.azcollection.azure

author:
    - Bill Peck (@p3ck)
'''

EXAMPLES = '''
- name: List ML Components
  azure.azcollection.azure_rm_ml_component_info:
    name: MyComponent
    resource_group: myResourceGroup
    ml_workspace: myMLWorkspace

- name: List ML Components
  azure.azcollection.azure_rm_ml_component_info:
    resource_group: myResourceGroup
    ml_workspace: myMLWorkspace
'''

RETURN = '''
ml_components:
    description:
        - Components that match the query.
    returned: always
    type: dict
    sample: [
      {
        "$schema": "https://azuremlschemas.azureedge.net/latest/commandComponent.schema.json",
        "code": "azureml:/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/xxxxx-ml-component/providers/Microsoft.MachineLearningServices/workspaces/workspace-xxxxxxxxxx/codes/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/versions/1",
        "command": "python train.py  --training_data ${{inputs.training_data}}  --max_epocs ${{inputs.max_epocs}}    --learning_rate ${{inputs.learning_rate}}  --learning_rate_schedule ${{inputs.learning_rate_schedule}}  --model_output ${{outputs.model_output}}",
        "creation_context": {
            "created_at": "2026-03-18T14:02:54.635544+00:00",
            "created_by": "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
            "created_by_type": "Application",
            "last_modified_at": "2026-03-18T14:02:54.773038+00:00",
            "last_modified_by": "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
            "last_modified_by_type": "Application"
        },
        "display_name": "Train_upper_case",
        "environment": "azureml://registries/azureml/environments/sklearn-1.5/versions/41",
        "id": "azureml:/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/xxxxx-ml-component/providers/Microsoft.MachineLearningServices/workspaces/workspace-xxxxxxxxxx/components/componentxxxxxxxxxx/labels/default",
        "inputs": {
            "learning_rate": {
                "default": "0.01",
                "optional": false,
                "type": "number"
            },
            "learning_rate_schedule": {
                "default": "time-based",
                "enum": [
                    "step",
                    "time-based"
                ],
                "optional": false,
                "type": "string"
            },
            "max_epocs": {
                "max": "100",
                "min": "0",
                "optional": false,
                "type": "integer"
            },
            "training_data": {
                "optional": false,
                "type": "uri_folder"
            }
        },
        "is_deterministic": true,
        "name": "componentxxxxxxxxxx",
        "outputs": {
            "model_output": {
                "type": "uri_folder"
            }
        },
        "resources": {
            "instance_count": 1
        },
        "type": "command",
        "version": "2"
      }
    ]
'''  # NOQA


try:
    from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common_ml import MLClientCommon
    from azure.core.exceptions import ResourceNotFoundError
except ImportError:
    # This is handled in azure_rm_common
    pass


class AzureRMMLComponentInfo(MLClientCommon):
    def __init__(self):

        self.module_arg_spec = dict(
            name=dict(
                type='str',
                required=False,
            ),
            resource_group=dict(
                type='str',
                required=True,
            ),
            ml_workspace=dict(
                type='str',
                required=True,
            ),
            ml_registry=dict(
                type='str',
            ),
            version=dict(
                type='str',
            ),
            label=dict(
                type='str',
            ),
            list_type=dict(
                type='str',
                choices=['archived', 'active', 'all'],
            ),
        )

        self._client = None

        self.results = dict(
            ml_components=[]
        )

        required_one_of = [('ml_workspace', 'ml_registry')]
        mutually_exclusive = [('version', 'label'),
                              ('ml_workspace', 'ml_registry'),
                              ('name', 'list_type')]
        required_together = [('ml_workspace', 'resource_group')]

        super(AzureRMMLComponentInfo, self).__init__(self.module_arg_spec,
                                                     supports_tags=False,
                                                     supports_check_mode=True,
                                                     facts_module=True,
                                                     required_one_of=required_one_of,
                                                     mutually_exclusive=mutually_exclusive,
                                                     required_together=required_together
                                                     )

    def exec_module(self, **kwargs):

        for key in list(self.module_arg_spec.keys()):
            setattr(self, key, kwargs[key])

        if self.name:
            try:
                result = self.client.components.get(name=self.name,
                                                    version=self.version,
                                                    label=self.label)
                ml_components = [self.entity_to_dict(result)]
            except ResourceNotFoundError:
                ml_components = []
        else:
            list_view_type = self.get_list_view_type(self.list_type)
            results = self.client.components.list(name=self.name,
                                                  list_view_type=list_view_type)
            ml_components = [self.entity_to_dict(self.client.components.get(x.name, version=x.latest_version)) for x in results]

        self.results['ml_components'] = ml_components
        return self.results


def main():
    AzureRMMLComponentInfo()


if __name__ == '__main__':
    main()



# This code is part of the Ansible collection community.docker, but is an independent component.
# This particular file, and this file only, is based on the Docker SDK for Python (https://github.com/docker/docker-py/)
#
# Copyright (c) 2016-2025 Docker, Inc.
#
# It is licensed under the Apache 2.0 license (see LICENSES/Apache-2.0.txt in this collection)
# SPDX-License-Identifier: Apache-2.0

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import json
import os
import typing as t

from .. import errors
from .config import (
    METAFILE,
    get_current_context_name,
    get_meta_dir,
    write_context_name_to_docker_config,
)
from .context import Context

if t.TYPE_CHECKING:
    from ..tls import TLSConfig


def create_default_context() -> Context:
    host = None
    if os.environ.get("DOCKER_HOST"):
        host = os.environ.get("DOCKER_HOST")
    return Context(
        "default", "swarm", host, description="Current DOCKER_HOST based configuration"
    )


class ContextAPI:
    """Context API.
    Contains methods for context management:
    create, list, remove, get, inspect.
    """

    DEFAULT_CONTEXT = None

    @classmethod
    def get_default_context(cls) -> Context:
        context = cls.DEFAULT_CONTEXT
        if context is None:
            context = create_default_context()
            cls.DEFAULT_CONTEXT = context
        return context

    @classmethod
    def create_context(
        cls,
        name: str,
        orchestrator: str | None = None,
        host: str | None = None,
        tls_cfg: TLSConfig | None = None,
        default_namespace: str | None = None,
        skip_tls_verify: bool = False,
    ) -> Context:
        """Creates a new context.
        Returns:
            (Context): a Context object.
        Raises:
            :py:class:`docker.errors.MissingContextParameter`
                If a context name is not provided.
            :py:class:`docker.errors.ContextAlreadyExists`
                If a context with the name already exists.
            :py:class:`docker.errors.ContextException`
                If name is default.

        Example:

        >>> from docker.context import ContextAPI
        >>> ctx = ContextAPI.create_context(name='test')
        >>> print(ctx.Metadata)
        {
            "Name": "test",
            "Metadata": {},
            "Endpoints": {
                "docker": {
                    "Host": "unix:///var/run/docker.sock",
                    "SkipTLSVerify": false
                }
            }
        }
        """
        if not name:
            raise errors.MissingContextParameter("name")
        if name == "default":
            raise errors.ContextException('"default" is a reserved context name')
        ctx = Context.load_context(name)
        if ctx:
            raise errors.ContextAlreadyExists(name)
        endpoint = "docker"
        if orchestrator and orchestrator != "swarm":
            endpoint = orchestrator
        ctx = Context(name, orchestrator)
        ctx.set_endpoint(
            endpoint,
            host,
            tls_cfg,
            skip_tls_verify=skip_tls_verify,
            def_namespace=default_namespace,
        )
        ctx.save()
        return ctx

    @classmethod
    def get_context(cls, name: str | None = None) -> Context | None:
        """Retrieves a context object.
        Args:
            name (str): The name of the context

        Example:

        >>> from docker.context import ContextAPI
        >>> ctx = ContextAPI.get_context(name='test')
        >>> print(ctx.Metadata)
        {
            "Name": "test",
            "Metadata": {},
            "Endpoints": {
                "docker": {
                "Host": "unix:///var/run/docker.sock",
                "SkipTLSVerify": false
                }
            }
        }
        """
        if not name:
            name = get_current_context_name()
        if name == "default":
            return cls.get_default_context()
        return Context.load_context(name)

    @classmethod
    def contexts(cls) -> list[Context]:
        """Context list.
        Returns:
            (Context): List of context objects.
        Raises:
            :py:class:`docker.errors.APIError`
                If something goes wrong.
        """
        names = []
        for dirname, dummy, fnames in os.walk(get_meta_dir()):
            for filename in fnames:
                if filename == METAFILE:
                    filepath = os.path.join(dirname, filename)
                    try:
                        with open(filepath, "rt", encoding="utf-8") as f:
                            data = json.load(f)
                        name = data["Name"]
                        if name == "default":
                            raise ValueError('"default" is a reserved context name')
                        names.append(name)
                    except Exception as e:
                        raise errors.ContextException(
                            f"Failed to load metafile {filepath}: {e}"
                        ) from e

        contexts = [cls.get_default_context()]
        for name in names:
            context = Context.load_context(name)
            if not context:
                raise errors.ContextException(f"Context {name} cannot be found")
            contexts.append(context)
        return contexts

    @classmethod
    def get_current_context(cls) -> Context | None:
        """Get current context.
        Returns:
            (Context): current context object.
        """
        return cls.get_context()

    @classmethod
    def set_current_context(cls, name: str = "default") -> None:
        ctx = cls.get_context(name)
        if not ctx:
            raise errors.ContextNotFound(name)

        err = write_context_name_to_docker_config(name)
        if err:
            raise errors.ContextException(f"Failed to set current context: {err}")

    @classmethod
    def remove_context(cls, name: str) -> None:
        """Remove a context. Similar to the ``docker context rm`` command.

        Args:
            name (str): The name of the context

        Raises:
            :py:class:`docker.errors.MissingContextParameter`
                If a context name is not provided.
            :py:class:`docker.errors.ContextNotFound`
                If a context with the name does not exist.
            :py:class:`docker.errors.ContextException`
                If name is default.

        Example:

        >>> from docker.context import ContextAPI
        >>> ContextAPI.remove_context(name='test')
        >>>
        """
        if not name:
            raise errors.MissingContextParameter("name")
        if name == "default":
            raise errors.ContextException('context "default" cannot be removed')
        ctx = Context.load_context(name)
        if not ctx:
            raise errors.ContextNotFound(name)
        if name == get_current_context_name():
            write_context_name_to_docker_config(None)
        ctx.remove()

    @classmethod
    def inspect_context(cls, name: str = "default") -> dict[str, t.Any]:
        """Inspect a context. Similar to the ``docker context inspect`` command.

        Args:
            name (str): The name of the context

        Raises:
            :py:class:`docker.errors.MissingContextParameter`
                If a context name is not provided.
            :py:class:`docker.errors.ContextNotFound`
                If a context with the name does not exist.

        Example:

        >>> from docker.context import ContextAPI
        >>> ContextAPI.remove_context(name='test')
        >>>
        """
        if not name:
            raise errors.MissingContextParameter("name")
        if name == "default":
            return cls.get_default_context()()
        ctx = Context.load_context(name)
        if not ctx:
            raise errors.ContextNotFound(name)

        return ctx()



#!/usr/bin/python

# Copyright (c) 2022, Håkon Lerring
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: consul_role
short_description: Manipulate Consul roles
version_added: 7.5.0
description:
  - Allows the addition, modification and deletion of roles in a Consul cluster using the agent. For more details on using
    and configuring ACLs, see U(https://www.consul.io/docs/guides/acl.html).
author:
  - Håkon Lerring (@Hakon)
extends_documentation_fragment:
  - community.general._consul
  - community.general._consul.token
  - community.general._consul.actiongroup_consul
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: partial
    details:
      - In check mode the diff misses operational attributes.
    version_added: 8.3.0
  action_group:
    version_added: 8.3.0
options:
  name:
    description:
      - A name used to identify the role.
    required: true
    type: str
  state:
    description:
      - Whether the role should be present or absent.
    choices: ['present', 'absent']
    default: present
    type: str
  description:
    description:
      - Description of the role.
      - If not specified, the assigned description is not changed.
    type: str
  policies:
    type: list
    elements: dict
    description:
      - List of policies to attach to the role. Each policy is a dict.
      - If the parameter is left blank, any policies currently assigned are not changed.
      - Any empty array (V([])) clears any policies previously set.
    suboptions:
      name:
        description:
          - The name of the policy to attach to this role; see M(community.general.consul_policy) for more info.
          - Either this or O(policies[].id) must be specified.
        type: str
      id:
        description:
          - The ID of the policy to attach to this role; see M(community.general.consul_policy) for more info.
          - Either this or O(policies[].name) must be specified.
        type: str
  templated_policies:
    description:
      - The list of templated policies that should be applied to the role.
    type: list
    elements: dict
    version_added: 8.3.0
    suboptions:
      template_name:
        description:
          - The templated policy name.
        type: str
        required: true
      template_variables:
        description:
          - The templated policy variables.
          - Not all templated policies require variables.
        type: dict
  service_identities:
    type: list
    elements: dict
    description:
      - List of service identities to attach to the role.
      - If not specified, any service identities currently assigned are not changed.
      - If the parameter is an empty array (V([])), any node identities assigned are unassigned.
    suboptions:
      service_name:
        description:
          - The name of the node.
          - Must not be longer than 256 characters, must start and end with a lowercase alphanumeric character.
          - May only contain lowercase alphanumeric characters as well as V(-) and V(_).
          - This suboption has been renamed from O(service_identities[].name) to O(service_identities[].service_name) in community.general
            8.3.0. The old name can still be used.
        type: str
        required: true
        aliases:
          - name
      datacenters:
        description:
          - The datacenters where the policies are effective.
          - This results in effective policy only being valid in this datacenter.
          - If an empty array (V([])) is specified, the policies are valid in all datacenters.
          - Including those which do not yet exist but may in the future.
        type: list
        elements: str
  node_identities:
    type: list
    elements: dict
    description:
      - List of node identities to attach to the role.
      - If not specified, any node identities currently assigned are not changed.
      - If the parameter is an empty array (V([])), any node identities assigned are unassigned.
    suboptions:
      node_name:
        description:
          - The name of the node.
          - Must not be longer than 256 characters, must start and end with a lowercase alphanumeric character.
          - May only contain lowercase alphanumeric characters as well as V(-) and V(_).
          - This suboption has been renamed from O(node_identities[].name) to O(node_identities[].node_name) in community.general
            8.3.0. The old name can still be used.
        type: str
        required: true
        aliases:
          - name
      datacenter:
        description:
          - The nodes datacenter.
          - This results in effective policy only being valid in this datacenter.
        type: str
        required: true
"""

EXAMPLES = r"""
- name: Create a role with 2 policies
  community.general.consul_role:
    host: consul1.example.com
    token: some_management_acl
    name: foo-role
    policies:
      - id: 783beef3-783f-f41f-7422-7087dc272765
      - name: "policy-1"

- name: Create a role with service identity
  community.general.consul_role:
    host: consul1.example.com
    token: some_management_acl
    name: foo-role-2
    service_identities:
      - name: web
        datacenters:
          - dc1

- name: Create a role with node identity
  community.general.consul_role:
    host: consul1.example.com
    token: some_management_acl
    name: foo-role-3
    node_identities:
      - name: node-1
        datacenter: dc2

- name: Remove a role
  community.general.consul_role:
    host: consul1.example.com
    token: some_management_acl
    name: foo-role-3
    state: absent
"""

RETURN = r"""
role:
  description: The role object.
  returned: success
  type: dict
  sample:
    {
      "CreateIndex": 39,
      "Description": "",
      "Hash": "Trt0QJtxVEfvTTIcdTUbIJRr6Dsi6E4EcwSFxx9tCYM=",
      "ID": "9a300b8d-48db-b720-8544-a37c0f5dafb5",
      "ModifyIndex": 39,
      "Name": "foo-role",
      "Policies": [
        {
          "ID": "b1a00172-d7a1-0e66-a12e-7a4045c4b774",
          "Name": "foo-access"
        }
      ]
    }
operation:
  description: The operation performed on the role.
  returned: changed
  type: str
  sample: update
"""

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._consul import (
    AUTH_ARGUMENTS_SPEC,
    OPERATION_READ,
    _ConsulModule,
)


class ConsulRoleModule(_ConsulModule):
    api_endpoint = "acl/role"
    result_key = "role"
    unique_identifiers = ["id"]

    def endpoint_url(self, operation, identifier=None):
        if operation == OPERATION_READ:
            return [self.api_endpoint, "name", self.params["name"]]
        return super().endpoint_url(operation, identifier)


NAME_ID_SPEC = dict(
    name=dict(type="str"),
    id=dict(type="str"),
)

NODE_ID_SPEC = dict(
    node_name=dict(type="str", required=True, aliases=["name"]),
    datacenter=dict(type="str", required=True),
)

SERVICE_ID_SPEC = dict(
    service_name=dict(type="str", required=True, aliases=["name"]),
    datacenters=dict(type="list", elements="str"),
)

TEMPLATE_POLICY_SPEC = dict(
    template_name=dict(type="str", required=True),
    template_variables=dict(type="dict"),
)

_ARGUMENT_SPEC = {
    "name": dict(type="str", required=True),
    "description": dict(type="str"),
    "policies": dict(
        type="list",
        elements="dict",
        options=NAME_ID_SPEC,
        mutually_exclusive=[("name", "id")],
        required_one_of=[("name", "id")],
    ),
    "templated_policies": dict(
        type="list",
        elements="dict",
        options=TEMPLATE_POLICY_SPEC,
    ),
    "node_identities": dict(
        type="list",
        elements="dict",
        options=NODE_ID_SPEC,
    ),
    "service_identities": dict(
        type="list",
        elements="dict",
        options=SERVICE_ID_SPEC,
    ),
    "state": dict(default="present", choices=["present", "absent"]),
}
_ARGUMENT_SPEC.update(AUTH_ARGUMENTS_SPEC)


def main():
    module = AnsibleModule(
        _ARGUMENT_SPEC,
        supports_check_mode=True,
    )
    consul_module = ConsulRoleModule(module)
    consul_module.execute()


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (c) 2017, 2018 Kairo Araujo <kairo@kairo.eti.br>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
author:
  - Kairo Araujo (@kairoaraujo)
module: aix_devices
short_description: Manages AIX devices
description:
  - This module discovers, defines, removes and modifies attributes of AIX devices.
deprecated:
  removed_in: 15.0.0
  why: The module is not actively maintained.
  alternative: >-
    Use C(ibm.power_aix.devices) instead.
    See U(https://ibm.github.io/ansible-power-aix/modules/devices.html) for details.
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  attributes:
    description:
      - A list of device attributes.
    type: dict
  device:
    description:
      - The name of the device.
      - V(all) is valid to rescan C(available) all devices (AIX C(cfgmgr) command).
    type: str
  force:
    description:
      - Forces action.
    type: bool
    default: false
  recursive:
    description:
      - Removes or defines a device and children devices.
    type: bool
    default: false
  state:
    description:
      - Controls the device state.
      - V(available) (alias V(present)) rescan a specific device or all devices (when O(device) is not specified).
      - V(removed) (alias V(absent) removes a device.
      - V(defined) changes device to Defined state.
    type: str
    choices: [available, defined, removed]
    default: available
"""

EXAMPLES = r"""
- name: Scan new devices
  community.general.aix_devices:
    device: all
    state: available

- name: Scan new virtual devices (vio0)
  community.general.aix_devices:
    device: vio0
    state: available

- name: Removing IP alias to en0
  community.general.aix_devices:
    device: en0
    attributes:
      delalias4: 10.0.0.100,255.255.255.0

- name: Removes ent2
  community.general.aix_devices:
    device: ent2
    state: removed

- name: Put device en2 in Defined
  community.general.aix_devices:
    device: en2
    state: defined

- name: Removes ent4 (inexistent).
  community.general.aix_devices:
    device: ent4
    state: removed

- name: Put device en4 in Defined (inexistent)
  community.general.aix_devices:
    device: en4
    state: defined

- name: Put vscsi1 and children devices in Defined state.
  community.general.aix_devices:
    device: vscsi1
    recursive: true
    state: defined

- name: Removes vscsi1 and children devices.
  community.general.aix_devices:
    device: vscsi1
    recursive: true
    state: removed

- name: Changes en1 mtu to 9000 and disables arp.
  community.general.aix_devices:
    device: en1
    attributes:
      mtu: 900
      arp: 'off'
    state: available

- name: Configure IP, netmask and set en1 up.
  community.general.aix_devices:
    device: en1
    attributes:
      netaddr: 192.168.0.100
      netmask: 255.255.255.0
      state: up
    state: available

- name: Adding IP alias to en0
  community.general.aix_devices:
    device: en0
    attributes:
      alias4: 10.0.0.100,255.255.255.0
    state: available
"""

RETURN = r""" # """

from ansible.module_utils.basic import AnsibleModule


def _check_device(module, device):
    """
    Check if device already exists and the state.
    Args:
        module: Ansible module.
        device: device to be checked.

    Returns: bool, device state

    """
    lsdev_cmd = module.get_bin_path("lsdev", True)
    rc, lsdev_out, err = module.run_command([lsdev_cmd, "-C", "-l", device])

    if rc != 0:
        module.fail_json(msg="Failed to run lsdev", rc=rc, err=err)

    if lsdev_out:
        device_state = lsdev_out.split()[1]
        return True, device_state

    device_state = None
    return False, device_state


def _check_device_attr(module, device, attr):
    """

    Args:
        module: Ansible module.
        device: device to check attributes.
        attr: attribute to be checked.

    Returns:

    """
    lsattr_cmd = module.get_bin_path("lsattr", True)
    rc, lsattr_out, err = module.run_command([lsattr_cmd, "-El", device, "-a", f"{attr}"])

    hidden_attrs = ["delalias4", "delalias6"]

    if rc == 255:
        if attr in hidden_attrs:
            current_param = ""
        else:
            current_param = None

        return current_param

    elif rc != 0:
        module.fail_json(msg=f"Failed to run lsattr: {err}", rc=rc, err=err)

    current_param = lsattr_out.split()[1]
    return current_param


def discover_device(module, device):
    """Discover AIX devices."""
    cfgmgr_cmd = module.get_bin_path("cfgmgr", True)

    if device is not None:
        device = f"-l {device}"

    else:
        device = ""

    changed = True
    msg = ""
    if not module.check_mode:
        rc, cfgmgr_out, err = module.run_command([cfgmgr_cmd, device])
        changed = True
        msg = cfgmgr_out

    return changed, msg


def change_device_attr(module, attributes, device, force):
    """Change AIX device attribute."""

    attr_changed = []
    attr_not_changed = []
    attr_invalid = []
    chdev_cmd = module.get_bin_path("chdev", True)

    for attr in list(attributes.keys()):
        new_param = attributes[attr]
        current_param = _check_device_attr(module, device, attr)

        if current_param is None:
            attr_invalid.append(attr)

        elif current_param != new_param:
            if force:
                cmd = [chdev_cmd, "-l", device, "-a", f"{attr}={attributes[attr]}", f"{force}"]
            else:
                cmd = [chdev_cmd, "-l", device, "-a", f"{attr}={attributes[attr]}"]

            if not module.check_mode:
                rc, chdev_out, err = module.run_command(cmd)
                if rc != 0:
                    module.fail_json(msg="Failed to run chdev.", rc=rc, err=err)

            attr_changed.append(attributes[attr])
        else:
            attr_not_changed.append(attributes[attr])

    if len(attr_changed) > 0:
        changed = True
        attr_changed_msg = f"Attributes changed: {','.join(attr_changed)}. "
    else:
        changed = False
        attr_changed_msg = ""

    if len(attr_not_changed) > 0:
        attr_not_changed_msg = f"Attributes already set: {','.join(attr_not_changed)}. "
    else:
        attr_not_changed_msg = ""

    if len(attr_invalid) > 0:
        attr_invalid_msg = f"Invalid attributes: {', '.join(attr_invalid)} "
    else:
        attr_invalid_msg = ""

    msg = f"{attr_changed_msg}{attr_not_changed_msg}{attr_invalid_msg}"

    return changed, msg


def remove_device(module, device, force, recursive, state):
    """Puts device in defined state or removes device."""

    state_opt = {"removed": "-d", "absent": "-d", "defined": ""}

    recursive_opt = {True: "-R", False: ""}

    recursive = recursive_opt[recursive]
    state = state_opt[state]

    changed = True
    msg = ""
    rmdev_cmd = module.get_bin_path("rmdev", True)

    if not module.check_mode:
        if state:
            rc, rmdev_out, err = module.run_command([rmdev_cmd, "-l", device, f"{recursive}", f"{force}"])
        else:
            rc, rmdev_out, err = module.run_command([rmdev_cmd, "-l", device, f"{recursive}"])

        if rc != 0:
            module.fail_json(msg="Failed to run rmdev", rc=rc, err=err)

        msg = rmdev_out

    return changed, msg


def main():
    module = AnsibleModule(
        argument_spec=dict(
            attributes=dict(type="dict"),
            device=dict(type="str"),
            force=dict(type="bool", default=False),
            recursive=dict(type="bool", default=False),
            state=dict(type="str", default="available", choices=["available", "defined", "removed"]),
        ),
        supports_check_mode=True,
    )

    force_opt = {
        True: "-f",
        False: "",
    }

    attributes = module.params["attributes"]
    device = module.params["device"]
    force = force_opt[module.params["force"]]
    recursive = module.params["recursive"]
    state = module.params["state"]

    result = dict(
        changed=False,
        msg="",
    )

    if state == "available" or state == "present":
        if attributes:
            # change attributes on device
            device_status, device_state = _check_device(module, device)
            if device_status:
                result["changed"], result["msg"] = change_device_attr(module, attributes, device, force)
            else:
                result["msg"] = f"Device {device} does not exist."

        else:
            # discovery devices (cfgmgr)
            if device and device != "all":
                device_status, device_state = _check_device(module, device)
                if device_status:
                    # run cfgmgr on specific device
                    result["changed"], result["msg"] = discover_device(module, device)

                else:
                    result["msg"] = f"Device {device} does not exist."

            else:
                result["changed"], result["msg"] = discover_device(module, device)

    elif state == "removed" or state == "absent" or state == "defined":
        if not device:
            result["msg"] = "device is required to removed or defined state."

        else:
            # Remove device
            check_device, device_state = _check_device(module, device)
            if check_device:
                if state == "defined" and device_state == "Defined":
                    result["changed"] = False
                    result["msg"] = f"Device {device} already in Defined"

                else:
                    result["changed"], result["msg"] = remove_device(module, device, force, recursive, state)

            else:
                result["msg"] = f"Device {device} does not exist."

    else:
        result["msg"] = f"Unexpected state {state}."
        module.fail_json(**result)

    module.exit_json(**result)


if __name__ == "__main__":
    main()



#!/usr/bin/python
#
# Copyright (c) 2018, Luca 'remix_tj' Lorenzetto <lorenzetto.luca@gmail.com>
#
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
#

from __future__ import annotations

DOCUMENTATION = r"""
module: emc_vnx_sg_member

short_description: Manage storage group member on EMC VNX


description:
  - This module manages the members of an existing storage group.
extends_documentation_fragment:
  - community.general._emc.emc_vnx
  - community.general._attributes

attributes:
  check_mode:
    support: full
  diff_mode:
    support: none

options:
  name:
    description:
      - Name of the Storage group to manage.
    required: true
    type: str
  lunid:
    description:
      - LUN ID to be added.
    required: true
    type: int
  state:
    description:
      - Indicates the desired lunid state.
      - V(present) ensures specified O(lunid) is present in the Storage Group.
      - V(absent) ensures specified O(lunid) is absent from Storage Group.
    default: present
    choices: ["present", "absent"]
    type: str


author:
  - Luca 'remix_tj' Lorenzetto (@remixtj)
"""

EXAMPLES = r"""
- name: Add lun to storage group
  community.general.emc_vnx_sg_member:
    name: sg01
    sp_address: sp1a.fqdn
    sp_user: sysadmin
    sp_password: sysadmin
    lunid: 100
    state: present

- name: Remove lun from storage group
  community.general.emc_vnx_sg_member:
    name: sg01
    sp_address: sp1a.fqdn
    sp_user: sysadmin
    sp_password: sysadmin
    lunid: 100
    state: absent
"""

RETURN = r"""
hluid:
  description: LUNID visible to hosts attached to the storage group.
  type: int
  returned: success
"""

import traceback

from ansible.module_utils.basic import AnsibleModule, missing_required_lib

from ansible_collections.community.general.plugins.module_utils._emc_vnx import emc_vnx_argument_spec

LIB_IMP_ERR = None
try:
    from storops import VNXSystem
    from storops.exception import (
        VNXAluAlreadyAttachedError,
        VNXAttachAluError,
        VNXCredentialError,
        VNXDetachAluNotFoundError,
        VNXStorageGroupError,
    )

    HAS_LIB = True
except Exception:
    LIB_IMP_ERR = traceback.format_exc()
    HAS_LIB = False


def run_module():
    module_args = dict(
        name=dict(type="str", required=True),
        lunid=dict(type="int", required=True),
        state=dict(default="present", choices=["present", "absent"]),
    )

    module_args.update(emc_vnx_argument_spec)

    result = dict(changed=False, hluid=None)

    module = AnsibleModule(argument_spec=module_args, supports_check_mode=True)

    if not HAS_LIB:
        module.fail_json(msg=missing_required_lib("storops >= 0.5.10"), exception=LIB_IMP_ERR)

    sp_user = module.params["sp_user"]
    sp_address = module.params["sp_address"]
    sp_password = module.params["sp_password"]
    alu = module.params["lunid"]

    # if the user is working with this module in only check mode we do not
    # want to make any changes to the environment, just return the current
    # state with no modifications
    if module.check_mode:
        return result

    try:
        vnx = VNXSystem(sp_address, sp_user, sp_password)
        sg = vnx.get_sg(module.params["name"])
        if sg.existed:
            if module.params["state"] == "present":
                if not sg.has_alu(alu):
                    try:
                        result["hluid"] = sg.attach_alu(alu)
                        result["changed"] = True
                    except VNXAluAlreadyAttachedError:
                        result["hluid"] = sg.get_hlu(alu)
                    except (VNXAttachAluError, VNXStorageGroupError) as e:
                        module.fail_json(msg=f"Error attaching {alu}: {e} ", **result)
                else:
                    result["hluid"] = sg.get_hlu(alu)
            if module.params["state"] == "absent" and sg.has_alu(alu):
                try:
                    sg.detach_alu(alu)
                    result["changed"] = True
                except VNXDetachAluNotFoundError:
                    # being not attached when using absent is OK
                    pass
                except VNXStorageGroupError as e:
                    module.fail_json(msg=f"Error detaching alu {alu}: {e} ", **result)
        else:
            module.fail_json(msg=f"No such storage group named {module.params['name']}", **result)
    except VNXCredentialError as e:
        module.fail_json(msg=f"{e}", **result)

    module.exit_json(**result)


def main():
    run_module()


if __name__ == "__main__":
    main()



# This code is part of Ansible, but is an independent component.
# This particular file snippet, and this file snippet only, is BSD licensed.
# Modules you write using this snippet, which is embedded dynamically by Ansible
# still belong to the author of the module, and may assign their own license
# to the complete work.
#
# (c) 2016 Red Hat Inc.
# (c) 2020 Cisco Systems Inc.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright notice,
#      this list of conditions and the following disclaimer in the documentation
#      and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# Intersight REST API Module
# Author: Matthew Garrett
# Contributors: David Soper, Chris Gascoigne, John McDonough

from __future__ import (absolute_import, division, print_function)

__metaclass__ = type

from base64 import b64encode
from email.utils import formatdate
import re
import json
import hashlib
from typing import Optional

from ansible.module_utils.six.moves.urllib.parse import urlparse, urlencode
from ansible.module_utils.urls import fetch_url
from ansible.module_utils.basic import env_fallback

try:
    from cryptography.hazmat.primitives import serialization, hashes
    from cryptography.hazmat.primitives.asymmetric import padding, ec, rsa
    from cryptography.hazmat.backends import default_backend

    HAS_CRYPTOGRAPHY = True
except ImportError:
    HAS_CRYPTOGRAPHY = False

intersight_argument_spec = dict(
    api_private_key=dict(fallback=(env_fallback, ['INTERSIGHT_API_PRIVATE_KEY']), type='path', no_log=True),
    api_uri=dict(fallback=(env_fallback, ['INTERSIGHT_API_URI']), type='str', default='https://intersight.com/api/v1'),
    api_key_id=dict(fallback=(env_fallback, ['INTERSIGHT_API_KEY_ID']), type='str'),
    api_bearer_token=dict(fallback=(env_fallback, ['INTERSIGHT_API_BEARER_TOKEN']), type='str', no_log=True),
    validate_certs=dict(type='bool', default=True),
    use_proxy=dict(type='bool', default=True),
)


def get_sha256_digest(data):
    """
    Generates a SHA256 digest from a String.

    :param data: data string set by user
    :return: instance of digest object
    """

    digest = hashlib.sha256()
    digest.update(data.encode())

    return digest


def prepare_str_to_sign(req_tgt, hdrs):
    """
    Concatenates Intersight headers in preparation to be signed

    :param req_tgt : http method plus endpoint
    :param hdrs: dict with header keys
    :return: concatenated header authorization string
    """
    ss = ""
    ss = ss + "(request-target): " + req_tgt + "\n"

    length = len(hdrs.items())

    i = 0
    for key, value in hdrs.items():
        ss = ss + key.lower() + ": " + value
        if i < length - 1:
            ss = ss + "\n"
        i += 1

    return ss


def get_gmt_date():
    """
    Generated a GMT formatted Date

    :return: current date
    """

    return formatdate(timeval=None, localtime=False, usegmt=True)


def compare_lists(expected_list, actual_list):
    if len(expected_list) != len(actual_list):
        # mismatch if list lengths aren't equal
        return False
    for expected, actual in zip(expected_list, actual_list):
        # if compare_values returns False, stop the loop and return
        if not compare_values(expected, actual):
            return False
    # loop complete with all items matching
    return True


def compare_values(expected, actual):
    try:
        if isinstance(expected, list) and isinstance(actual, list):
            return compare_lists(expected, actual)
        for (key, value) in expected.items():
            if re.search(r'P(ass)?w(or)?d', key) or re.search(r'OctetString', key) or re.search(r'IsOctetStringSet', key) or key not in actual:
                # do not compare any password related attributes, OctetString (secrets), IsOctetStringSet, or attributes that are not in the actual resource
                continue
            # Special handling for SecKeys comparison
            if key == 'SecKeys':
                expected_empty = value is None or value == []
                actual_value = actual.get(key)
                actual_empty = actual_value is None or actual_value == [] or actual_value == ''
                if expected_empty and actual_empty:
                    continue  # Both are "empty", consider them equal
                # If SecKeys has content, skip deep comparison (contains secrets we can't compare)
                # Just verify both have content
                if not expected_empty and not actual_empty:
                    continue  # Both have keys, consider them equal (can't compare secrets)
            if not compare_values(value, actual[key]):
                return False
        # loop complete with all items matching
        return True
    except (AttributeError, TypeError):
        # Handle Intersight object references: when expected is a MOID string and actual is an object reference
        if (isinstance(expected, str) and isinstance(actual, dict) and 'Moid' in actual and actual.get('ClassId') == 'mo.MoRef'):
            # Compare the MOID string against the Moid field in the object reference
            if actual['Moid'] == expected:
                return True
        # Handle case where expected is a string and actual is an object reference
        elif (isinstance(expected, str) and isinstance(actual, dict) and
              'Name' in actual and actual.get('ObjectType') == 'organization.Organization'):
            if actual['Name'] == expected:
                return True
        # if expected and actual != expected:
        if actual != expected:
            return False
        return True


# This will fix the format of a v3 secret key, if needed.
# Some v3 keys PEM files are incorrectly formatted with
# "BEGIN EC PRIVATE KEY" instead of "BEGIN PRIVATE KEY"
def _fix_v3_key_format(secret_key: bytes):
    return secret_key.replace(
        b"-----BEGIN EC PRIVATE KEY-----", b"-----BEGIN PRIVATE KEY-----"
    ).replace(b"-----END EC PRIVATE KEY-----", b"-----END PRIVATE KEY-----")


class IntersightModule():

    def __init__(self, module):
        self.module = module
        self.result = dict(changed=False)
        self.host = self.module.params['api_uri']
        self.bearer_token = self.module.params.get('api_bearer_token')
        self.digest_algorithm = ''
        self.response_list = []
        self.update_method = ''

        if self.bearer_token:
            self.public_key = None
            self.private_key = None
        else:
            if not self.module.params.get('api_key_id'):
                self.module.fail_json(
                    msg='api_key_id is required when api_bearer_token is not provided'
                )
            if not self.module.params.get('api_private_key'):
                self.module.fail_json(
                    msg='api_private_key is required when api_bearer_token is not provided'
                )
            if not HAS_CRYPTOGRAPHY:
                self.module.fail_json(msg='cryptography is required for this module')
            self.public_key = self.module.params['api_key_id']
            try:
                with open(self.module.params['api_private_key'], 'r') as f:
                    self.private_key = f.read()
            except (FileNotFoundError, OSError):
                self.private_key = self.module.params['api_private_key']

    def get_sig_b64encode(self, data):
        """
        Generates a signed digest from a String

        :param digest: string to be signed & hashed
        :return: instance of digest object
        """
        try:
            key = serialization.load_pem_private_key(self.private_key.encode(), None, default_backend())
        except (ValueError):
            key = serialization.load_pem_private_key(_fix_v3_key_format(self.private_key.encode()), None, default_backend())

        if isinstance(key, rsa.RSAPrivateKey):
            sign = key.sign(data.encode(), padding.PKCS1v15(), hashes.SHA256())
            self.digest_algorithm = 'rsa-sha256'
        elif isinstance(key, ec.EllipticCurvePrivateKey):
            sign = key.sign(data.encode(), ec.ECDSA(hashes.SHA256()))
            self.digest_algorithm = 'hs2019'
        else:
            raise Exception("Unsupported key: {0}".format(type(key).__name__))

        return b64encode(sign)

    def get_auth_header(self, hdrs, signed_msg):
        """
        Assmebled an Intersight formatted authorization header

        :param hdrs : object with header keys
        :param signed_msg: base64 encoded sha256 hashed body
        :return: concatenated authorization header
        """

        auth_str = "Signature"

        auth_str = auth_str + " " + "keyId=\"" + self.public_key + "\"," + "algorithm=\"" + self.digest_algorithm + "\","

        auth_str = auth_str + "headers=\"(request-target)"

        for key, dummy in hdrs.items():
            auth_str = auth_str + " " + key.lower()
        auth_str = auth_str + "\""

        auth_str = auth_str + "," + "signature=\"" + signed_msg.decode('ascii') + "\""

        return auth_str

    def call_api(self, **options):
        """
        Call the Intersight API and check for success status
        :param options: options dict with method and other params for API call
        :return: json http response object
        """

        try:
            response, info = self.intersight_call(**options)
            if not re.match(r'2..', str(info['status'])):
                raise RuntimeError(info.get('status'), info.get('msg'), info.get('body', info))
        except Exception as e:
            self.module.fail_json(msg="Code exception: %s " % str(e))

        response_data = response.read()
        if len(response_data) > 0:
            resp_json = json.loads(response_data)
            resp_json['trace_id'] = info.get('x-starship-traceid')
            return resp_json
        return {}

    def intersight_call(self, http_method="", resource_path="", query_params=None, body=None, moid=None, name=None):
        """
        Invoke the Intersight API

        :param resource_path: intersight resource path e.g. '/ntp/Policies'
        :param query_params: dictionary object with query string parameters as key/value pairs
        :param body: dictionary object with intersight data
        :param moid: intersight object moid
        :param name: intersight object name
        :return: json http response object
        """

        target_host = urlparse(self.host).netloc
        target_path = urlparse(self.host).path
        query_path = ""
        method = http_method.upper()
        bodyString = ""

        # Verify an accepted HTTP verb was chosen
        if (method not in ['GET', 'POST', 'PATCH', 'DELETE']):
            raise ValueError('Please select a valid HTTP verb (GET/POST/PATCH/DELETE)')

        # Verify the resource path isn't empy & is a valid <str> object
        if (resource_path != "" and not (resource_path, str)):
            raise TypeError('The *resource_path* value is required and must be of type "<str>"')

        # Verify the query parameters isn't empy & is a valid <dict> object
        if (query_params is not None and not isinstance(query_params, dict)):
            raise TypeError('The *query_params* value must be of type "<dict>"')

        # Verify the MOID is not null & of proper length
        if (moid is not None and len(moid.encode('utf-8')) != 24):
            raise ValueError('Invalid *moid* value!')

        if (method != 'PATCH' and self.update_method == 'json-patch'):
            raise ValueError('json-patch is only supported with PATCH on existing resource')

        # Check for query_params, encode, and concatenate onto URL
        if query_params:
            query_path = "?" + urlencode(query_params)

        # Handle PATCH/DELETE by Object "name" instead of "moid"
        if method in ('PATCH', 'DELETE'):
            if moid is None:
                if name is not None:
                    if isinstance(name, str):
                        moid = self.get_moid_by_name(resource_path, name)
                    else:
                        raise TypeError('The *name* value must be of type "<str>"')
                else:
                    raise ValueError('Must set either *moid* or *name* with "PATCH/DELETE!"')

        # Check for moid and concatenate onto URL
        if moid is not None:
            resource_path += "/" + moid

        # Check for GET request to properly form body
        if method != "GET":
            bodyString = json.dumps(body)

        # Concatenate URLs for headers
        target_url = self.host + resource_path + query_path
        request_target = method.lower() + " " + target_path + resource_path + query_path

        # Get the current GMT Date/Time
        cdate = get_gmt_date()

        if self.update_method == 'json-patch':
            content_type = 'application/json-patch+json'
        else:
            content_type = 'application/json'

        if self.bearer_token:
            request_header = {
                'Accept': 'application/json',
                'Content-Type': content_type,
                'Host': '{0}'.format(target_host),
                'Date': '{0}'.format(cdate),
                'Authorization': 'Bearer {0}'.format(self.bearer_token),
            }
        else:
            # Generate the body digest
            body_digest = get_sha256_digest(bodyString)
            b64_body_digest = b64encode(body_digest.digest())

            # Generate the authorization header
            auth_header = {
                'Host': target_host,
                'Date': cdate,
                'Digest': "SHA-256=" + b64_body_digest.decode('ascii'),
            }

            string_to_sign = prepare_str_to_sign(request_target, auth_header)
            b64_signed_msg = self.get_sig_b64encode(string_to_sign)
            auth_header = self.get_auth_header(auth_header, b64_signed_msg)

            request_header = {
                'Accept': 'application/json',
                'Content-Type': content_type,
                'Host': '{0}'.format(target_host),
                'Date': '{0}'.format(cdate),
                'Digest': 'SHA-256={0}'.format(b64_body_digest.decode('ascii')),
                'Authorization': '{0}'.format(auth_header),
            }

        response, info = fetch_url(self.module, target_url, data=bodyString, headers=request_header, method=method,
                                   use_proxy=self.module.params['use_proxy'])

        return response, info

    def get_resource(self, resource_path, query_params, return_list=False):
        '''
        GET a resource and return the 1st element found or the full Results list
        If return_list is False and more than 1 element is returned, a warning is raised
        '''

        if query_params is None:
            query_params = {}
        # Handle pagination, default return list consists of 100 elements.
        if return_list and '$top' not in query_params:
            query_params['$top'] = 1000
        options = {
            'http_method': 'get',
            'resource_path': resource_path,
            'query_params': query_params,
        }
        response = self.call_api(**options)
        if response.get('Results'):
            if return_list:
                self.result['api_response'] = response['Results']
            else:
                if len(response['Results']) > 1:
                    self.module.warn('More than 1 resource found, returning the 1st one')
                # return the 1st list element
                self.result['api_response'] = response['Results'][0]
        else:
            # Return a dict if no Results array (handles case where resource_path contains a Moid)
            if isinstance(response, dict):
                self.result['api_response'] = response
            else:
                # Clear api_response when no results found to prevent returning stale data
                if return_list:
                    self.result['api_response'] = []
                else:
                    self.result['api_response'] = {}
        self.result['count'] = response.get('Count', 0)
        self.result['trace_id'] = response.get('trace_id')

    def configure_resource(self, moid, resource_path, body, query_params, update_method=''):
        self.update_method = update_method
        if not self.module.check_mode:
            if moid and update_method != 'post':
                # update the resource - user has to specify all the props they want updated
                options = {
                    'http_method': 'patch',
                    'resource_path': resource_path,
                    'body': body,
                    'moid': moid,
                }
                response_dict = self.call_api(**options)
                if response_dict.get('Results'):
                    # return the 1st element in the results list
                    self.result['api_response'] = response_dict['Results'][0]
                    self.result['trace_id'] = response_dict.get('trace_id')
                elif response_dict and 'Moid' in response_dict:
                    # PATCH returned the updated object directly (not in Results array)
                    self.result['api_response'] = response_dict
                    self.result['trace_id'] = response_dict.get('trace_id')
            else:
                # create the resource
                options = {
                    'http_method': 'post',
                    'resource_path': resource_path,
                    'body': body,
                }
                response_dict = self.call_api(**options)
                if response_dict:
                    self.result['api_response'] = response_dict
                    self.result['trace_id'] = response_dict.get('trace_id')
                elif query_params:
                    # POSTs may not return any data.
                    # Get the current state of the resource if query_params.
                    self.get_resource(
                        resource_path=resource_path,
                        query_params=query_params,
                    )
        else:
            # In check mode, clear api_response to indicate no actual action was taken
            self.result['api_response'] = {}
        self.result['changed'] = True

    def delete_resource(self, moid, resource_path):
        # delete resource and create empty api_response
        if not self.module.check_mode:
            options = {
                'http_method': 'delete',
                'resource_path': resource_path,
                'moid': moid,
            }
            resp = self.call_api(**options)
            self.result['api_response'] = {}
            self.result['trace_id'] = resp.get('trace_id')
        self.result['changed'] = True

    def configure_policy_or_profile(self, resource_path, filter_key=None, filter_value=None):
        # Configure (create, update, or delete) the policy or profile
        organization_name = self.module.params['organization']
        organization_moid = self.get_moid_by_name(resource_path='/organization/Organizations', resource_name=organization_name)

        if organization_moid is None:
            self.module.fail_json(
                msg=f"Organization '{organization_name}' not found. Please verify the organization name exists in Intersight and that you have access to it."
            )

        self.result['api_response'] = {}
        # Get the current state of the resource
        filter_str = "Name eq '" + self.module.params['name'] + "'"
        if filter_value and filter_key:
            filter_str += " and " + filter_key + " eq '" + filter_value + "'"
        filter_str += " and Organization.Moid eq '" + organization_moid + "'"
        self.get_resource(
            resource_path=resource_path,
            query_params={
                '$filter': filter_str,
                '$expand': 'Organization',
            }
        )

        moid = None
        resource_values_match = False
        if self.result['api_response'].get('Moid'):
            # resource exists and moid was returned
            moid = self.result['api_response']['Moid']
            if self.module.params['state'] == 'present':
                resource_values_match = compare_values(self.api_body, self.result['api_response'])
            else:  # state == 'absent'
                self.delete_resource(
                    moid=moid,
                    resource_path=resource_path,
                )
                moid = None

        if self.module.params['state'] == 'present' and not resource_values_match:
            # remove read-only Organization key
            self.api_body.pop('Organization')
            if not moid:
                # Organization must be set, but can't be changed after initial POST
                self.api_body['Organization'] = {
                    'Moid': organization_moid,
                }
            self.configure_resource(
                moid=moid,
                resource_path=resource_path,
                body=self.api_body,
                query_params={
                    '$filter': filter_str
                }
            )
            if self.result['api_response'].get('Moid'):
                # resource exists and moid was returned
                moid = self.result['api_response']['Moid']

        return moid

    def configure_secondary_resource(self, resource_path, resource_name=None, state='present', custom_filter=None):
        # Configure (create, update, or delete) resources
        # This method is used to configure secondery resources that are part of a policy or profile (e.g. VLANs)

        self.result['api_response'] = {}
        # Get the current state of the resource
        if custom_filter:
            filter_str = custom_filter
        elif resource_name:
            filter_str = "Name eq '" + resource_name + "'"
        else:
            raise ValueError("Either resource_name or custom_filter must be provided")
        self.get_resource(
            resource_path=resource_path,
            query_params={
                '$filter': filter_str
            }
        )

        moid = None
        resource_values_match = False
        if self.result['api_response'].get('Moid'):
            # resource exists and moid was returned
            moid = self.result['api_response']['Moid']
            if state == 'present':
                resource_values_match = compare_values(self.api_body, self.result['api_response'])
            else:  # state == 'absent'
                self.delete_resource(
                    moid=moid,
                    resource_path=resource_path,
                )
                moid = None

        if state == 'present' and not resource_values_match:
            self.configure_resource(
                moid=moid,
                resource_path=resource_path,
                body=self.api_body,
                query_params={
                    '$filter': filter_str
                }
            )
            if self.result['api_response'].get('Moid'):
                # resource exists and moid was returned
                moid = self.result['api_response']['Moid']

        return moid

    def get_moid_by_name(self, resource_path, resource_name) -> Optional[str]:
        '''
        Get the moid of the resource by the value of the resource
        '''
        organization_moid = None
        # GET Moid of the resource
        self.get_resource(
            resource_path=resource_path,
            query_params={
                '$filter': "Name eq '" + resource_name + "'",
                '$select': 'Moid',
            },
        )
        if self.result['api_response'].get('Moid'):
            # resource exists and moid was returned
            return self.result['api_response']['Moid']
        return None

    def get_moid_by_name_and_org(self, resource_path, resource_name, organization_name) -> Optional[str]:
        '''
        Get the moid of an organization-scoped resource by name and organization.

        This function is used for resources that belong to a specific organization
        to avoid conflicts when multiple organizations have resources with the same name.
        '''
        # Get the organization MOID
        org_moid = self.get_moid_by_name(
            resource_path='/organization/Organizations',
            resource_name=organization_name
        )

        if not org_moid:
            self.module.fail_json(msg=f"Organization '{organization_name}' not found")

        # Build filter with both name and organization
        filter_str = f"Name eq '{resource_name}' and Organization.Moid eq '{org_moid}'"

        # GET Moid of the resource
        self.get_resource(
            resource_path=resource_path,
            query_params={
                '$filter': filter_str,
                '$select': 'Moid',
            },
        )

        if self.result['api_response'].get('Moid'):
            # resource exists and moid was returned
            return self.result['api_response']['Moid']
        return None

    def set_query_params(self, filter_key=None, filter_value=None) -> dict:
        filter_conditions = []
        name_to_filter = self.module.params.get('name')
        org_to_filter = self.module.params.get('organization')
        if name_to_filter:
            filter_conditions.append(f"Name eq '{name_to_filter}'")

        if org_to_filter:
            org_moid = self.get_moid_by_name(resource_path='/organization/Organizations', resource_name=org_to_filter)
            filter_conditions.append(f"Organization.Moid eq '{org_moid}'")

        if filter_value and filter_key:
            filter_conditions.append(f"{filter_key} eq '{filter_value}'")

        query_params = {}
        if filter_conditions:
            query_params["$filter"] = " and ".join(filter_conditions)
        return query_params

    def set_tags_and_description(self):
        """
        Generalize the pattern of setting tags and description in api_body if they are provided in module params.
        """
        if self.module.params.get('tags'):
            self.api_body['Tags'] = self.module.params['tags']
        if self.module.params.get('description'):
            self.api_body['Description'] = self.module.params['description']

    def execute_bulk_operations(self, resource_path, operations_dict, changed_states):
        """
        Execute bulk operations (POST, PATCH, DELETE) and track results.

        This is a generic method to execute categorized bulk operations and collect results.
        It handles all three operation types and properly tracks changed states.
        Order of operations does not affect the final changed state - any operation that
        makes changes will result in changed=True at the end.

        Args:
            resource_path: API resource path (e.g., '/fabric/Vlans')
            operations_dict: Dictionary with keys 'create', 'update', 'delete', each containing
                           a list of operation data dicts
            changed_states: List to append changed state to after each operation

        Returns:
            List of all operation results from create and update operations
        """
        all_results = []

        # Execute bulk CREATE operations
        if operations_dict.get('create'):
            # Reset changed flag before operation to ensure clean state tracking
            self.result['changed'] = False
            create_results = self.configure_bulk_resources(
                resource_path=resource_path,
                resources_data=operations_dict['create'],
                http_method='post'
            )
            all_results.extend(create_results)
            # Track if this operation made changes
            changed_states.append(self.result['changed'])

        # Execute bulk UPDATE operations
        if operations_dict.get('update'):
            # Reset changed flag before operation to ensure clean state tracking
            self.result['changed'] = False
            update_results = self.configure_bulk_resources(
                resource_path=resource_path,
                resources_data=operations_dict['update'],
                http_method='patch'
            )
            all_results.extend(update_results)
            # Track if this operation made changes
            changed_states.append(self.result['changed'])

        # Execute bulk DELETE operations
        if operations_dict.get('delete'):
            # Reset changed flag before operation to ensure clean state tracking
            self.result['changed'] = False
            self.configure_bulk_resources(
                resource_path=resource_path,
                resources_data=operations_dict['delete'],
                http_method='delete'
            )
            # Track if this operation made changes (no results to return for deletes)
            changed_states.append(self.result['changed'])

        return all_results

    def configure_bulk_resources(self, resource_path, resources_data, http_method='post'):
        """
        Configure multiple resources in bulk operations with automatic chunking.

        This method uses the Intersight bulk API endpoint to create, update, or delete multiple
        resources. It automatically splits large operations into chunks of 100 requests (Intersight API limit)
        and processes them sequentially.

        Args:
            resource_path: API resource path (e.g., '/fabric/Vlans')
            resources_data: List of dicts, each containing:
                - 'body': Resource API body (required for POST/PATCH)
                - 'moid': Resource MOID (required for PATCH/DELETE)
                - 'filter': Filter string to check existence (optional)
                - 'name': Resource name for identification (optional)
            http_method: 'post' for create, 'patch' for update, or 'delete' for delete operations

        Returns:
            List of API response objects for each resource in the bulk operation

        Raises:
            ValueError: If http_method is not 'post', 'patch', or 'delete'
        """
        if http_method.lower() not in ['post', 'patch', 'delete']:
            raise ValueError("Bulk operations only support 'post' (create), 'patch' (update), and 'delete' methods")

        if not resources_data:
            return []

        # Intersight API limit: maximum 100 requests per bulk operation
        BULK_CHUNK_SIZE = 100
        bulk_results = []

        # Split resources_data into chunks of 100
        for chunk_start in range(0, len(resources_data), BULK_CHUNK_SIZE):
            # Prevent index out of range error
            chunk_end = min(chunk_start + BULK_CHUNK_SIZE, len(resources_data))
            # Get the chunk of resources data
            chunk_data = resources_data[chunk_start:chunk_end]

            # Process this chunk based on HTTP method
            if http_method.lower() == 'post':
                chunk_results = self._process_bulk_post_chunk(resource_path, chunk_data)
            elif http_method.lower() == 'patch':
                chunk_results = self._process_bulk_patch_chunk(resource_path, chunk_data)
            elif http_method.lower() == 'delete':
                chunk_results = self._process_bulk_delete_chunk(resource_path, chunk_data)

            bulk_results.extend(chunk_results)

        return bulk_results

    def _process_bulk_response(self, response, chunk_data, operation_type, success_statuses=None):
        """
        Process bulk API response and collect failures.

        Args:
            response: API response dictionary
            chunk_data: Original chunk data for resource identification
            operation_type: Type of operation (POST, PATCH, DELETE)
            success_statuses: List of HTTP status codes considered successful (defaults to [200])

        Returns:
            List of bulk operation results
        """
        if success_statuses is None:
            success_statuses = [200]

        bulk_results = []
        failed_count = 0
        failure_messages = []

        if response.get('Results'):
            for idx, result in enumerate(response['Results']):
                if result.get('Status') in success_statuses:
                    bulk_results.append(result.get('Body', {}))
                else:
                    failed_count += 1
                    error_body = result.get('Body', {})
                    error_msg = error_body.get('Message', error_body.get('message', 'Unknown error'))
                    resource_name = chunk_data[idx].get('name', f'resource {idx}')
                    failure_msg = f"  - {resource_name}: {error_msg}"
                    if failure_msg not in failure_messages:
                        failure_messages.append(failure_msg)
                    bulk_results.append({})

            # Handle any failures
            self._handle_bulk_failures(failed_count, len(response['Results']), failure_messages, operation_type)
        else:
            # If no Results, return empty results for each item
            bulk_results = [{}] * len(chunk_data)

        return bulk_results

    def _handle_bulk_failures(self, failed_count, total_count, failure_messages, operation_type):
        """
        Handle bulk operation failures with consistent error reporting.

        Args:
            failed_count: Number of failed operations
            total_count: Total number of operations attempted
            failure_messages: List of error messages
            operation_type: Type of operation (POST, PATCH, DELETE)
        """
        if failed_count > 0:
            error_summary = "\n".join(failure_messages[:10])  # Show up to 10 errors
            self.module.fail_json(
                msg=f"Bulk {operation_type} operation failed: {failed_count}/{total_count} resources failed. "
                    f"Errors:\n{error_summary}"
            )

    def _process_bulk_post_chunk(self, resource_path, chunk_data):
        """
        Process a chunk of POST (create) operations.

        Args:
            resource_path: API resource path
            chunk_data: List of resources to create (max 100 items)

        Returns:
            List of API response objects for created resources
        """
        bulk_results = []

        # Build bulk request body
        bulk_requests = []
        for resource_data in chunk_data:
            bulk_requests.append({
                'ObjectType': 'bulk.RestSubRequest',
                'Body': resource_data['body']
            })

        bulk_body = {
            'Verb': 'POST',
            'Uri': f'/v1{resource_path}',
            'Requests': bulk_requests
        }

        if not self.module.check_mode:
            # Make bulk API call
            options = {
                'http_method': 'post',
                'resource_path': '/bulk/Requests',
                'body': bulk_body
            }
            response = self.call_api(**options)
            bulk_results = self._process_bulk_response(response, chunk_data, 'POST')
        else:
            # In check mode, return empty results
            bulk_results = [{}] * len(chunk_data)

        self.result['changed'] = True
        return bulk_results

    def _process_bulk_patch_chunk(self, resource_path, chunk_data):
        """
        Process a chunk of PATCH (update) operations.

        Args:
            resource_path: API resource path
            chunk_data: List of resources to update (max 100 items)

        Returns:
            List of API response objects for updated resources
        """
        bulk_results = []

        # Build bulk request body
        bulk_requests = []
        for resource_data in chunk_data:
            if 'moid' in resource_data and resource_data['moid']:
                bulk_requests.append({
                    'ObjectType': 'bulk.RestSubRequest',
                    'TargetMoid': resource_data['moid'],
                    'Body': resource_data['body']
                })

        if bulk_requests:
            bulk_body = {
                'Verb': 'PATCH',
                'Uri': f'/v1{resource_path}',
                'Requests': bulk_requests
            }

            if not self.module.check_mode:
                # Make bulk API call
                options = {
                    'http_method': 'post',
                    'resource_path': '/bulk/Requests',
                    'body': bulk_body
                }
                response = self.call_api(**options)
                bulk_results = self._process_bulk_response(response, chunk_data, 'PATCH')
            else:
                # In check mode, return empty results
                bulk_results = [{}] * len(bulk_requests)

            self.result['changed'] = True

        return bulk_results

    def _process_bulk_delete_chunk(self, resource_path, chunk_data):
        """
        Process a chunk of DELETE operations.

        Args:
            resource_path: API resource path
            chunk_data: List of resources to delete (max 100 items)

        Returns:
            List of empty dicts (delete operations don't return data)
        """
        bulk_results = []

        # Build bulk request body
        bulk_requests = []
        for resource_data in chunk_data:
            if 'moid' in resource_data and resource_data['moid']:
                sub_request = {
                    'ObjectType': 'bulk.RestSubRequest',
                    'TargetMoid': resource_data['moid']
                }
                # Include any additional fields from resource_data (e.g., VlanId for VLANs)
                if 'body' in resource_data and resource_data['body']:
                    sub_request.update(resource_data['body'])
                bulk_requests.append(sub_request)

        if bulk_requests:
            bulk_body = {
                'Verb': 'DELETE',
                'Uri': f'/v1{resource_path}',
                'Requests': bulk_requests
            }

            if not self.module.check_mode:
                # Make bulk API call
                options = {
                    'http_method': 'post',
                    'resource_path': '/bulk/Requests',
                    'body': bulk_body
                }
                response = self.call_api(**options)
                bulk_results = self._process_bulk_response(response, chunk_data, 'DELETE', success_statuses=[200, 204])
            else:
                # In check mode, return empty results
                bulk_results = [{}] * len(bulk_requests)

            self.result['changed'] = True

        return bulk_results


def resolve_policy_bucket(intersight, organization_name, policy_mapping):
    """Resolve policy names to MOIDs and build the PolicyBucket list.

    Args:
        intersight: IntersightModule instance.
        organization_name: Organization to scope the policy lookup.
        policy_mapping: Dict mapping module param names to resource_path/object_type.
    """
    policy_bucket = []
    for param_name, mapping in policy_mapping.items():
        policy_name = intersight.module.params.get(param_name)
        if policy_name:
            moid = intersight.get_moid_by_name_and_org(
                resource_path=mapping['resource_path'],
                resource_name=policy_name,
                organization_name=organization_name,
            )
            if not moid:
                intersight.module.fail_json(
                    msg=f"Policy '{policy_name}' ({param_name}) not found in organization '{organization_name}'"
                )
            policy_bucket.append({
                'Moid': moid,
                'ObjectType': mapping['object_type'],
            })
    return policy_bucket


def sync_policy_bucket(intersight, bucket_path, desired_bucket, current_bucket):
    """Sync the PolicyBucket via the sub-resource endpoint using differential add/delete.

    Args:
        intersight: IntersightModule instance.
        bucket_path: Full API sub-resource path (e.g. /chassis/Profiles/{moid}/PolicyBucket).
        desired_bucket: List of policy dicts representing the desired state.
        current_bucket: List of policy dicts representing the current state.
    """
    current_set = {(p['Moid'], p['ObjectType']) for p in current_bucket}
    desired_set = {(p['Moid'], p['ObjectType']) for p in desired_bucket}

    if current_set == desired_set:
        return

    to_remove = current_set - desired_set
    for policy_moid, obj_type in to_remove:
        intersight.delete_resource(
            moid=policy_moid,
            resource_path=bucket_path,
        )

    to_add = desired_set - current_set
    if to_add:
        add_body = [{'Moid': m, 'ObjectType': t} for m, t in to_add]
        intersight.configure_resource(
            moid=None,
            resource_path=bucket_path,
            body=add_body,
            query_params={},
        )



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: ec2_carrier_gateway_info
version_added: 6.0.0
short_description: Gather information about carrier gateways in AWS
description:
  - Gather information about carrier gateways in AWS.
author:
  - "Marco Braga (@mtulio)"
options:
  filters:
    description:
      - A dict of filters to apply. Each dict item consists of a filter key and a filter value.
        See U(https://docs.aws.amazon.com/AWSEC2/latest/APIReference/API_DescribeCarrierGateways.html) for possible filters.
    required: false
    default: {}
    type: dict
  carrier_gateway_ids:
    description:
      - Get details of specific Carrier Gateway ID.
    required: false
    type: list
    elements: str
extends_documentation_fragment:
  - amazon.aws.common.modules
  - amazon.aws.region.modules
  - amazon.aws.boto3
"""

EXAMPLES = r"""
# # Note: These examples do not set authentication details, see the AWS Guide for details.

- name: Gather information about all Carrier Gateways for an account or profile
  community.aws.ec2_carrier_gateway_info:
    region: ap-southeast-2
  register: cagw_info

- name: Gather information about a filtered list of Carrier Gateways
  community.aws.ec2_carrier_gateway_info:
    region: ap-southeast-2
    filters:
      "tag:Name": "cagw-123"
  register: cagw_info

- name: Gather information about a specific carrier gateway by CarrierGatewayId
  community.aws.ec2_carrier_gateway_info:
    region: ap-southeast-2
    carrier_gateway_ids: cagw-c1231234
  register: cagw_info
"""

RETURN = r"""
changed:
    description: True if listing the carrier gateways succeeds.
    type: bool
    returned: always
    sample: "false"
carrier_gateways:
    description: The carrier gateways for the account.
    returned: always
    type: complex
    contains:
        vpc_id:
            description: The ID of the VPC.
            returned: I(state=present)
            type: str
            sample: vpc-02123b67
        carrier_gateway_id:
            description: The ID of the carrier gateway.
            returned: I(state=present)
            type: str
            sample: cagw-2123634d
        tags:
            description: Any tags assigned to the carrier gateway.
            returned: I(state=present)
            type: dict
            sample:
                tags:
                    "Ansible": "Test"
"""

try:
    import botocore
except ImportError:
    pass  # Handled by AnsibleAWSModule

from ansible.module_utils.common.dict_transformations import camel_dict_to_snake_dict

from ansible_collections.amazon.aws.plugins.module_utils.botocore import is_boto3_error_code
from ansible_collections.amazon.aws.plugins.module_utils.retries import AWSRetry
from ansible_collections.amazon.aws.plugins.module_utils.tagging import boto3_tag_list_to_ansible_dict
from ansible_collections.amazon.aws.plugins.module_utils.transformation import ansible_dict_to_boto3_filter_list

from ansible_collections.community.aws.plugins.module_utils.modules import AnsibleCommunityAWSModule as AnsibleAWSModule


def get_carrier_gateway_info(carrier_gateway):
    tags = boto3_tag_list_to_ansible_dict(carrier_gateway["Tags"])
    ignore_list = []
    carrier_gateway_info = {
        "CarrierGatewayId": carrier_gateway["CarrierGatewayId"],
        "VpcId": carrier_gateway["VpcId"],
        "Tags": tags,
    }

    carrier_gateway_info = camel_dict_to_snake_dict(carrier_gateway_info, ignore_list=ignore_list)
    return carrier_gateway_info


def list_carrier_gateways(connection, module):
    params = dict()

    params["Filters"] = ansible_dict_to_boto3_filter_list(module.params.get("filters"))
    if module.params.get("carrier_gateway_ids"):
        params["CarrierGatewayIds"] = module.params.get("carrier_gateway_ids")

    try:
        all_carrier_gateways = connection.describe_carrier_gateways(aws_retry=True, **params)
    except is_boto3_error_code("InvalidCarrierGatewayID.NotFound"):
        module.fail_json("CarrierGateway not found")
    except (
        botocore.exceptions.ClientError,
        botocore.exceptions.BotoCoreError,
    ) as e:  # pylint: disable=duplicate-except
        module.fail_json_aws(e, "Unable to describe carrier gateways")

    return [get_carrier_gateway_info(cagw) for cagw in all_carrier_gateways["CarrierGateways"]]


def main():
    argument_spec = dict(
        carrier_gateway_ids=dict(default=None, elements="str", type="list"),
        filters=dict(default={}, type="dict"),
    )

    module = AnsibleAWSModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
    )

    # Validate Requirements
    try:
        connection = module.client("ec2", retry_decorator=AWSRetry.jittered_backoff())
    except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
        module.fail_json_aws(e, msg="Failed to connect to AWS")

    results = list_carrier_gateways(connection, module)

    module.exit_json(carrier_gateways=results)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}

DOCUMENTATION = r'''
---
module: intersight_snmp_policy
short_description: SNMP Policy configuration for Cisco Intersight
description:
  - Manages SNMP Policy configuration on Cisco Intersight.
  - A policy to configure SNMP settings for Cisco Intersight managed servers.
  - Supports both SNMPv2c and SNMPv3 configurations with users, traps, and community strings.
  - For more information see L(Cisco Intersight,https://intersight.com/apidocs/snmp/Policy/get/).
extends_documentation_fragment: intersight
options:
  state:
    description:
      - If C(present), will verify the resource is present and will create if needed.
      - If C(absent), will verify the resource is absent and will delete if needed.
    type: str
    choices: [present, absent]
    default: present
  organization:
    description:
      - The name of the Organization this resource is assigned to.
      - Profiles, Policies, and Pools that are created within a Custom Organization are applicable only to devices in the same Organization.
    type: str
    default: default
  name:
    description:
      - The name assigned to the SNMP Policy.
      - The name must be between 1 and 64 alphanumeric characters, allowing special characters :-_.
    type: str
    required: true
  description:
    description:
      - The user-defined description for the SNMP Policy.
      - Description can contain letters(a-z, A-Z), numbers(0-9), hyphen(-), period(.), colon(:), or an underscore(_).
    type: str
    aliases: [descr]
  tags:
    description:
      - List of tags in Key:<user-defined key> Value:<user-defined value> format.
    type: list
    elements: dict
  enabled:
    description:
      - State of the SNMP Policy on the endpoint.
      - If enabled, the endpoint sends SNMP traps to the designated host.
    type: bool
    default: true
  v2c_enabled:
    description:
      - State of the SNMPv2c protocol.
      - When enabled, SNMPv2c access is available for the associated servers.
    type: bool
    default: true
  v3_enabled:
    description:
      - State of the SNMPv3 protocol.
      - When enabled, SNMPv3 access is available for the associated servers.
    type: bool
    default: true
  snmp_port:
    description:
      - Port on which Cisco IMC SNMP agent runs. Enter a value between 1-65535.
      - Reserved ports not allowed (22, 23, 80, 123, 389, 443, 623, 636, 2068, 3268, 3269).
    type: int
    default: 161
  sys_contact:
    description:
      - Contact person responsible for the SNMP implementation.
      - Enter a string up to 64 characters, such as an email address or a name and telephone number.
      - Required when C(enabled) is C(true).
    type: str
  sys_location:
    description:
      - Location of the host on which the SNMP agent (server) runs.
      - Required when C(enabled) is C(true).
    type: str
  community_access:
    description:
      - Controls access to the information in the inventory tables.
      - Applicable only for SNMPv2c users.
      - For SNMPv3, this is always set to C(Disabled).
      - For SNMPv2c, valid choices are C(Disabled), C(Limited), and C(Full).
      - Required when C(enabled) is C(true).
    type: str
    choices: [Disabled, Limited, Full]
    default: Disabled
  access_community_string:
    description:
      - The default SNMPv1, SNMPv2c community name or SNMPv3 username to include on any trap messages sent to the SNMP host.
      - The name can be 32 characters long.
      - Used with SNMPv2c access only.
    type: str
  trap_community:
    description:
      - SNMP community group used for sending SNMP trap to other devices.
      - Valid only for SNMPv2c users.
    type: str
  engine_input_id:
    description:
      - User-defined unique identification of the static engine.
      - Used with SNMPv3 only.
    type: str
  snmp_users:
    description:
      - List of SNMP users for SNMPv3 authentication.
    type: list
    elements: dict
    suboptions:
      name:
        description:
          - SNMP username.
          - Must have a minimum of 1 and and a maximum of 31 characters.
        type: str
        required: true
      security_level:
        description:
          - Security mechanism used for communication between agent and manager.
        type: str
        choices: [AuthPriv, AuthNoPriv]
        default: AuthPriv
      auth_password:
        description:
          - Authorization password for the user.
          - Required when security_level is AuthPriv or AuthNoPriv.
        type: str
        required: true
      privacy_password:
        description:
          - Privacy password for the SNMP user.
          - Required when security_level is AuthPriv.
        type: str
  snmp_traps:
    description:
      - List of SNMP trap destinations.
    type: list
    elements: dict
    suboptions:
      enabled:
        description:
          - Enables/disables the trap on the server If enabled, trap is active on the server.
        type: bool
        default: true
      version:
        description:
          - SNMP version for the trap.
        type: str
        choices: [V2, V3]
        required: true
      community:
        description:
          - SNMP community group used for sending SNMP trap to other devices.
          - Applicable only for SNMP v2c.
        type: str
      user:
        description:
          - SNMP user for the trap.
          - Applicable only to SNMPv3.
        type: str
      type:
        description:
          - Type of trap which decides whether to receive a notification when a trap is received at the destination.
          - Note that 'Inform' is only supported for V2 traps, V3 traps only support 'Trap' type.
        type: str
        choices: [Trap, Inform]
        default: Trap
      destination:
        description:
          - IP address or hostname of the trap destination.
        type: str
        required: true
      port:
        description:
          - Port number for the trap destination.
        type: int
        default: 162
author:
  - Ron Gershburg (@rgershbu)
'''

EXAMPLES = r'''
- name: Create SNMP Policy with SNMPv2 and SNMPv3 enabled
  cisco.intersight.intersight_snmp_policy:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    organization: "default"
    name: "mixed-snmp-policy"
    description: "SNMP policy with both v2 and v3 enabled"
    enabled: true
    v2c_enabled: true
    v3_enabled: true
    snmp_port: 161
    sys_contact: "admin@example.com"
    sys_location: "Data Center A"
    community_access: "Full"
    access_community_string: "public"
    trap_community: "trapcomm"
    engine_input_id: "custom-engine-id"
    snmp_users:
      - name: "admin"
        security_level: "AuthPriv"
        auth_password: "authpassword123"
        privacy_password: "privpassword123"
      - name: "readonly"
        security_level: "AuthNoPriv"
        auth_password: "readonlypass123"
    snmp_traps:
      - enabled: true
        version: "V2"
        community: "trapcomm"
        type: "Trap"
        destination: "192.168.1.100"
        port: 162
      - enabled: true
        version: "V3"
        user: "admin"
        type: "Trap"
        destination: "192.168.1.101"
        port: 162
    tags:
      - Key: "Environment"
        Value: "Production"
    state: present

- name: Create SNMP Policy with only SNMPv3 enabled
  cisco.intersight.intersight_snmp_policy:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    name: "v3-only-snmp-policy"
    description: "SNMP policy with only v3 enabled"
    enabled: true
    v2c_enabled: false
    v3_enabled: true
    snmp_port: 161
    sys_contact: "admin@example.com"
    sys_location: "Data Center B"
    engine_input_id: "v3-engine-id"
    snmp_users:
      - name: "v3user"
        security_level: "AuthPriv"
        auth_password: "v3authpass123"
        privacy_password: "v3privpass123"
    state: present

- name: Create SNMP Policy with only SNMPv2 enabled
  cisco.intersight.intersight_snmp_policy:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    name: "v2-only-snmp-policy"
    description: "SNMP policy with only v2 enabled"
    enabled: true
    v2c_enabled: true
    v3_enabled: false
    snmp_port: 161
    sys_contact: "admin@example.com"
    sys_location: "Data Center C"
    community_access: "Limited"
    access_community_string: "readonly"
    trap_community: "v2traps"
    state: present

- name: Create disabled SNMP Policy
  cisco.intersight.intersight_snmp_policy:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    name: "disabled-snmp-policy"
    description: "Disabled SNMP policy"
    enabled: false
    state: present

- name: Delete SNMP Policy
  cisco.intersight.intersight_snmp_policy:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    name: "mixed-snmp-policy"
    state: absent
'''

RETURN = r'''
api_response:
  description: The API response output returned by the specified resource.
  returned: always
  type: dict
  sample:
    "api_response": {
        "Name": "mixed-snmp-policy",
        "ObjectType": "snmp.Policy",
        "Enabled": true,
        "V2Enabled": true,
        "V3Enabled": true,
        "SnmpPort": 161,
        "SysContact": "admin@example.com",
        "SysLocation": "Data Center A",
        "CommunityAccess": "Full",
        "AccessCommunityString": "public",
        "TrapCommunity": "trapcomm",
        "EngineId": "custom-engine-id",
        "SnmpUsers": [
            {
                "Name": "admin",
                "SecurityLevel": "AuthPriv",
                "AuthType": "SHA",
                "IsAuthPasswordSet": true,
                "PrivacyType": "AES",
                "IsPrivacyPasswordSet": true
            }
        ],
        "SnmpTraps": [
            {
                "Enabled": true,
                "Version": "V2",
                "Community": "trapcomm",
                "Type": "Trap",
                "Destination": "192.168.1.100",
                "Port": 162
            }
        ],
        "Tags": [
            {
                "Key": "Environment",
                "Value": "Production"
            }
        ]
    }
'''

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.intersight.plugins.module_utils.intersight import IntersightModule, intersight_argument_spec


def get_argument_spec():
    snmp_user_spec = dict(
        name=dict(type='str', required=True),
        security_level=dict(type='str', choices=['AuthPriv', 'AuthNoPriv'], default='AuthPriv'),
        auth_password=dict(type='str', no_log=True, required=True),
        privacy_password=dict(type='str', no_log=True)
    )

    snmp_trap_spec = dict(
        enabled=dict(type='bool', default=True),
        version=dict(type='str', choices=['V2', 'V3'], required=True),
        community=dict(type='str'),
        user=dict(type='str'),
        type=dict(type='str', choices=['Trap', 'Inform'], default='Trap'),
        destination=dict(type='str', required=True),
        port=dict(type='int', default=162)
    )

    argument_spec = intersight_argument_spec.copy()
    argument_spec.update(
        state=dict(type='str', choices=['present', 'absent'], default='present'),
        organization=dict(type='str', default='default'),
        name=dict(type='str', required=True),
        description=dict(type='str', aliases=['descr']),
        tags=dict(type='list', elements='dict'),
        enabled=dict(type='bool', default=True),
        v2c_enabled=dict(type='bool', default=True),
        v3_enabled=dict(type='bool', default=True),
        snmp_port=dict(type='int', default=161),
        sys_contact=dict(type='str'),
        sys_location=dict(type='str'),
        community_access=dict(type='str', choices=['Disabled', 'Limited', 'Full'], default='Disabled'),
        access_community_string=dict(type='str'),
        trap_community=dict(type='str'),
        engine_input_id=dict(type='str'),
        snmp_users=dict(type='list', elements='dict', options=snmp_user_spec),
        snmp_traps=dict(type='list', elements='dict', options=snmp_trap_spec)
    )
    return argument_spec


def validate_snmp_configuration(module):
    """
    Validate SNMP configuration parameters.
    """
    params = module.params
    if params['state'] == 'present':
        # If SNMP is disabled, we don't need to validate other parameters
        if not params['enabled']:
            return

        # If SNMP is enabled, at least one version must be enabled
        if not params['v2c_enabled'] and not params['v3_enabled']:
            module.fail_json(msg="At least one of v2c_enabled or v3_enabled must be true when SNMP is enabled")

        # Check if both sys_contact and sys_location are provided when SNMP is enabled
        if not params.get('sys_contact') or not params.get('sys_location'):
            module.fail_json(msg="sys_contact and sys_location are required when SNMP is enabled")

        # SNMPv3 users can only be specified when SNMPv3 is enabled
        if not params['v3_enabled'] and params.get('snmp_users'):
            module.fail_json(msg="snmp_users cannot be specified when v3_enabled is false")

        # SNMPv3 only specific validations
        if params['v3_enabled'] and not params['v2c_enabled']:
            # For SNMPv3, community access is always Disabled
            if params['community_access'] != 'Disabled':
                module.fail_json(msg="community_access must be 'Disabled' when v3_enabled is true and v2c_enabled is false")

        # Validate SNMP users
        if params.get('snmp_users'):
            for user in params['snmp_users']:
                if not user.get('auth_password'):
                    module.fail_json(msg="auth_password is required for security_level 'AuthPriv' or 'AuthNoPriv'")

                if user['security_level'] == 'AuthPriv' and not user.get('privacy_password'):
                    module.fail_json(msg="privacy_password is required for security_level 'AuthPriv'")

        # Validate SNMP traps
        if params.get('snmp_traps'):
            for trap in params['snmp_traps']:
                if trap['version'] == 'V2' and not params['v2c_enabled']:
                    module.fail_json(msg="SNMP trap version 'V2' is not supported when v2c_enabled is false")
                elif trap['version'] == 'V3' and not params['v3_enabled']:
                    module.fail_json(msg="SNMP trap version 'V3' is not supported when v3_enabled is false")

                if trap['version'] == 'V2' and not trap.get('community'):
                    module.fail_json(msg="community is required for SNMP trap version 'V2'")
                elif trap['version'] == 'V3' and not trap.get('user'):
                    module.fail_json(msg="user is required for SNMP trap version 'V3'")

                # V3 traps only support 'Trap' type, not 'Inform'
                if trap['version'] == 'V3' and trap.get('type') == 'Inform':
                    module.fail_json(msg="SNMP trap version 'V3' does not support type 'Inform', only 'Trap' is supported")


def build_snmp_users(snmp_users):
    """
    Build SNMP users list for API body.

    Args:
        snmp_users: List of SNMP user dictionaries

    Returns:
        list: List of SNMP user dictionaries formatted for API
    """
    if not snmp_users:
        return []

    users = []
    for user in snmp_users:
        snmp_user = {
            'Name': user['name'],
            'SecurityLevel': user['security_level'],
            'AuthType': 'SHA',
            'AuthPassword': user['auth_password']
        }

        # Only add privacy type and password if security level is AuthPriv
        if user.get('security_level') == 'AuthPriv':
            snmp_user['PrivacyType'] = 'AES'
            snmp_user['PrivacyPassword'] = user['privacy_password']

        users.append(snmp_user)

    return users


def build_snmp_traps(snmp_traps):
    """
    Build SNMP traps list for API body.

    Args:
        snmp_traps: List of SNMP trap dictionaries

    Returns:
        list: List of SNMP trap dictionaries formatted for API
    """
    if not snmp_traps:
        return []

    traps = []
    for trap in snmp_traps:
        snmp_trap = {
            'Enabled': trap['enabled'],
            'Version': trap['version'],
            'Type': trap['type'],
            'Destination': trap['destination'],
            'Port': trap['port'],
        }

        if trap['version'] == 'V3':
            snmp_trap['User'] = trap['user']
        elif trap['version'] == 'V2':
            snmp_trap['Community'] = trap['community']

        traps.append(snmp_trap)

    return traps


def build_api_body(intersight):
    """
    Build the API body for SNMP policy configuration.

    Args:
        intersight: IntersightModule instance

    Returns:
        None: Updates intersight.api_body directly
    """
    params = intersight.module.params
    if params['state'] == 'present':
        # Base API body
        intersight.api_body = {
            'Organization': {
                'Name': params['organization'],
            },
            'Name': params['name'],
            'Enabled': params['enabled']
        }

        # Only include additional parameters if SNMP is enabled
        if params['enabled']:
            # Add version-specific parameters
            intersight.api_body['V2Enabled'] = params.get('v2c_enabled', True)
            intersight.api_body['V3Enabled'] = params.get('v3_enabled', True)

            # Add SNMP port
            intersight.api_body['SnmpPort'] = params['snmp_port']

            # Add system contact and location
            intersight.api_body['SysContact'] = params['sys_contact']
            intersight.api_body['SysLocation'] = params['sys_location']

            # Add community access
            intersight.api_body['CommunityAccess'] = params['community_access']

            # Add SNMPv2c specific parameters
            if params['v2c_enabled']:
                if params.get('access_community_string'):
                    intersight.api_body['AccessCommunityString'] = params['access_community_string']
                if params.get('trap_community'):
                    intersight.api_body['TrapCommunity'] = params['trap_community']

            # Add SNMPv3 specific parameters
            if params['v3_enabled']:
                if params.get('engine_input_id'):
                    intersight.api_body['EngineId'] = params['engine_input_id']

            # Add SNMP users
            intersight.api_body['SnmpUsers'] = build_snmp_users(params.get('snmp_users'))

            # Add SNMP traps
            intersight.api_body['SnmpTraps'] = build_snmp_traps(params.get('snmp_traps'))

        else:
            # When SNMP is disabled, send empty arrays
            intersight.api_body['SnmpUsers'] = []
            intersight.api_body['SnmpTraps'] = []

        # Add optional parameters
        intersight.set_tags_and_description()


def main():
    # Get the argument specification
    argument_spec = get_argument_spec()

    # Create the module
    module = AnsibleModule(
        argument_spec,
        supports_check_mode=True,
    )

    # Validate SNMP configuration
    validate_snmp_configuration(module)

    # Initialize Intersight module
    intersight = IntersightModule(module)
    intersight.result['api_response'] = {}
    intersight.result['trace_id'] = ''

    # Build API body
    build_api_body(intersight)

    # Resource path used to configure policy
    resource_path = '/snmp/Policies'

    # Configure the policy
    intersight.configure_policy_or_profile(resource_path=resource_path)

    # Exit with results
    module.exit_json(**intersight.result)


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2026, Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

DOCUMENTATION = r'''
---
module: mysql_perf_schema

short_description: Manage MySQL or MariaDB Performance Schema setup tables

description:
  - Manage runtime Performance Schema configuration through setup tables.
  - Supports instruments, consumers, actors, and objects.
  - Reconciles only the rows requested in the task and leaves unrelated rows untouched.
  - This module manages runtime state only. Restart-persistent Performance Schema configuration remains outside its scope.

author:
  - Ron Gershburg (@ronger4)

version_added: '5.1.0'

options:
  instruments:
    description:
      - Instrument rows to reconcile in C(performance_schema.setup_instruments).
      - Existing instrument rows are updated in place.
    type: list
    elements: dict
    suboptions:
      name:
        description:
          - Instrument name or pattern from C(performance_schema.setup_instruments), for example C(statement/%).
        type: str
        required: true
      enabled:
        description:
          - Whether the instrument should be enabled.
        type: bool
        required: true
      timed:
        description:
          - Whether the instrument should be timed.
        type: bool
        required: true
  consumers:
    description:
      - Consumer rows to reconcile in C(performance_schema.setup_consumers).
      - Existing consumer rows are updated in place.
    type: list
    elements: dict
    suboptions:
      name:
        description:
          - Consumer name.
        type: str
        required: true
      enabled:
        description:
          - Whether the consumer should be enabled.
        type: bool
        required: true
  actors:
    description:
      - Actor rows to reconcile in C(performance_schema.setup_actors).
    type: list
    elements: dict
    suboptions:
      user:
        description:
          - Account user pattern.
        type: str
        required: true
      host:
        description:
          - Account host pattern.
        type: str
        required: true
      role:
        description:
          - Role pattern.
        type: str
        required: true
      enabled:
        description:
          - Whether matching threads should be instrumented.
          - Required when C(state=present).
        type: bool
      history:
        description:
          - Whether history consumers should retain data for matching threads.
          - Required when C(state=present).
        type: bool
      state:
        description:
          - Whether the actor row should exist.
        type: str
        choices: [absent, present]
        default: present
  objects:
    description:
      - Object rows to reconcile in C(performance_schema.setup_objects).
    type: list
    elements: dict
    suboptions:
      object_type:
        description:
          - Object type stored in Performance Schema.
        type: str
        required: true
      object_schema:
        description:
          - Schema name pattern.
        type: str
        required: true
      object_name:
        description:
          - Object name pattern.
        type: str
        required: true
      enabled:
        description:
          - Whether the object should be instrumented.
          - Required when C(state=present).
        type: bool
      timed:
        description:
          - Whether the object should be timed.
          - Required when C(state=present).
        type: bool
      state:
        description:
          - Whether the object row should exist.
        type: str
        choices: [absent, present]
        default: present

notes:
  - Compatible with MariaDB or MySQL when the requested Performance Schema sections are available on the server.
  - The module requires Performance Schema to be enabled and the requested setup table columns to exist.
  - Changes are runtime only and are not persisted across restart by this module.

attributes:
  check_mode:
    support: full
  idempotent:
    support: full

seealso:
  - module: ansible.mysql.mysql_query
  - name: MySQL Performance Schema reference
    description: Oracle MySQL Performance Schema documentation.
    link: https://dev.mysql.com/doc/en/performance-schema.html
  - name: MariaDB Performance Schema overview
    description: MariaDB Performance Schema documentation.
    link: https://mariadb.com/docs/server/reference/system-tables/performance-schema/performance-schema-overview

extends_documentation_fragment:
  - ansible.mysql.mysql
'''

EXAMPLES = r'''
- name: Apply a standard monitoring baseline profile
  ansible.mysql.mysql_perf_schema:
    login_unix_socket: /run/mysqld/mysqld.sock
    instruments:
      - name: statement/sql/select
        enabled: true
        timed: true
      - name: stage/sql/Opening_tables
        enabled: true
        timed: true
      - name: wait/io/table/sql/handler
        enabled: true
        timed: true
    consumers:
      - name: events_statements_history
        enabled: true
      - name: events_waits_current
        enabled: true

- name: Enable statement instrumentation for one instrument
  ansible.mysql.mysql_perf_schema:
    login_unix_socket: /run/mysqld/mysqld.sock
    instruments:
      - name: statement/sql/insert
        enabled: true
        timed: true

- name: Enable one consumer and add one actor row
  ansible.mysql.mysql_perf_schema:
    login_unix_socket: /run/mysqld/mysqld.sock
    consumers:
      - name: events_waits_current
        enabled: true
    actors:
      - user: app
        host: '%'
        role: '%'
        enabled: true
        history: false

- name: Remove one object rule
  ansible.mysql.mysql_perf_schema:
    login_unix_socket: /run/mysqld/mysqld.sock
    objects:
      - object_type: TABLE
        object_schema: reporting
        object_name: slow_queries
        state: absent
'''

RETURN = r'''
queries:
  description: List of executed SQL statements or predicted SQL statements in check mode.
  returned: when changed
  type: list
  sample:
    - "UPDATE performance_schema.setup_consumers SET ENABLED = 'YES' WHERE NAME = 'events_waits_current'"
instruments:
  description: Normalized requested instrument rows after module execution or prediction.
  returned: when O(instruments) is provided
  type: list
  elements: dict
  sample:
    - name: statement/sql/select
      enabled: true
      timed: true
consumers:
  description: Normalized requested consumer rows after module execution or prediction.
  returned: when O(consumers) is provided
  type: list
  elements: dict
  sample:
    - name: events_waits_current
      enabled: true
actors:
  description: Normalized requested actor rows after module execution or prediction.
  returned: when O(actors) is provided
  type: list
  elements: dict
  sample:
    - user: app
      host: '%'
      role: '%'
      enabled: true
      history: false
objects:
  description: Normalized requested object rows after module execution or prediction.
  returned: when O(objects) is provided
  type: list
  elements: dict
  sample:
    - object_type: TABLE
      object_schema: reporting
      object_name: slow_queries
      enabled: true
      timed: false
'''

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.common.text.converters import to_native

from ansible_collections.ansible.mysql.plugins.module_utils.mysql import (
    mysql_common_argument_spec,
    mysql_connect,
    mysql_driver,
    mysql_driver_fail_msg,
)
from ansible_collections.ansible.mysql.plugins.module_utils.perf_schema import (
    SECTION_DEFINITIONS,
    ensure_perf_schema_sections_supported,
    plan_section_changes,
)


class MySQLPerfSchema(object):
    def __init__(self, module, cursor):
        self.module = module
        self.cursor = cursor

    def apply(self, params):
        sections = [section for section in SECTION_DEFINITIONS if params.get(section)]
        ensure_perf_schema_sections_supported(self.module, self.cursor, sections)

        result = {
            'changed': False,
            'queries': [],
        }

        for section in sections:
            current_rows = self.get_section_rows(section)
            planned = plan_section_changes(section, params[section], current_rows)
            result['changed'] = result['changed'] or planned['changed']
            result['queries'].extend(query['display'] for query in planned['queries'])
            result[section] = planned['rows']

            if not self.module.check_mode:
                self.execute_queries(planned['queries'])

        if not result['queries']:
            del result['queries']

        return result

    def get_section_rows(self, section):
        definition = SECTION_DEFINITIONS[section]
        fields = ', '.join(definition['db_key_fields'] + definition['db_value_fields'])
        query = 'SELECT %s FROM performance_schema.%s' % (fields, definition['table'])
        self.cursor.execute(query)
        return self.cursor.fetchall()

    def execute_queries(self, queries):
        for query in queries:
            self.cursor.execute(query['sql'], query['params'])
            self.cursor.fetchall()


def main():
    argument_spec = mysql_common_argument_spec()
    argument_spec.update(
        instruments=dict(
            type='list',
            elements='dict',
            options=dict(
                name=dict(type='str', required=True),
                enabled=dict(type='bool', required=True),
                timed=dict(type='bool', required=True),
            ),
        ),
        consumers=dict(
            type='list',
            elements='dict',
            options=dict(
                name=dict(type='str', required=True),
                enabled=dict(type='bool', required=True),
            ),
        ),
        actors=dict(
            type='list',
            elements='dict',
            required_if=[('state', 'present', ('enabled', 'history'))],
            options=dict(
                user=dict(type='str', required=True),
                host=dict(type='str', required=True),
                role=dict(type='str', required=True),
                enabled=dict(type='bool'),
                history=dict(type='bool'),
                state=dict(type='str', choices=['absent', 'present'], default='present'),
            ),
        ),
        objects=dict(
            type='list',
            elements='dict',
            required_if=[('state', 'present', ('enabled', 'timed'))],
            options=dict(
                object_type=dict(type='str', required=True),
                object_schema=dict(type='str', required=True),
                object_name=dict(type='str', required=True),
                enabled=dict(type='bool'),
                timed=dict(type='bool'),
                state=dict(type='str', choices=['absent', 'present'], default='present'),
            ),
        ),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        required_one_of=[('instruments', 'consumers', 'actors', 'objects')],
        supports_check_mode=True,
    )

    if mysql_driver is None:
        module.fail_json(msg=mysql_driver_fail_msg)

    try:
        cursor, db_conn = mysql_connect(
            module,
            module.params['login_user'],
            module.params['login_password'],
            module.params['config_file'],
            module.params['client_cert'],
            module.params['client_key'],
            module.params['ca_cert'],
            connect_timeout=module.params['connect_timeout'],
            check_hostname=module.params['check_hostname'],
            cursor_class='DictCursor',
            autocommit=True,
        )
    except Exception as e:
        module.fail_json(msg='unable to connect to database: %s' % to_native(e))

    executor = MySQLPerfSchema(module, cursor)

    try:
        result = executor.apply(module.params)
    except ValueError as e:
        module.fail_json(msg='invalid performance schema request: %s' % to_native(e))
    except Exception as e:
        module.fail_json(msg=to_native(e))

    module.exit_json(**result)


if __name__ == '__main__':
    main()



#!/usr/bin/python

# Copyright (c) 2017-18, Abhijeet Kasurde <akasurde@redhat.com>
#
#  GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: github_issue
short_description: View GitHub issue
description:
  - View GitHub issue for a given repository and organization.
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  repo:
    description:
      - Name of repository from which issue needs to be retrieved.
    required: true
    type: str
  organization:
    description:
      - Name of the GitHub organization in which the repository is hosted.
    required: true
    type: str
  issue:
    description:
      - Issue number for which information is required.
    required: true
    type: int
  action:
    description:
      - Get various details about issue depending upon action specified.
    default: 'get_status'
    choices:
      - get_status
    type: str
author:
  - Abhijeet Kasurde (@Akasurde)
"""

RETURN = r"""
issue_status:
  description: State of the GitHub issue.
  type: str
  returned: success
  sample: open, closed
"""

EXAMPLES = r"""
- name: Check if GitHub issue is closed or not
  community.general.github_issue:
    organization: ansible
    repo: ansible
    issue: 23642
    action: get_status
  register: r

- name: Take action depending upon issue status
  ansible.builtin.debug:
    msg: Do something when issue 23642 is open
  when: r.issue_status == 'open'
"""

import json

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.urls import fetch_url


def main():
    module = AnsibleModule(
        argument_spec=dict(
            organization=dict(required=True),
            repo=dict(required=True),
            issue=dict(type="int", required=True),
            action=dict(choices=["get_status"], default="get_status"),
        ),
        supports_check_mode=True,
    )

    organization = module.params["organization"]
    repo = module.params["repo"]
    issue = module.params["issue"]
    action = module.params["action"]

    result = dict()

    headers = {
        "Content-Type": "application/json",
        "Accept": "application/vnd.github.v3+json",
    }

    url = f"https://api.github.com/repos/{organization}/{repo}/issues/{issue}"

    response, info = fetch_url(module, url, headers=headers)
    if not (200 <= info["status"] < 400):
        if info["status"] == 404:
            module.fail_json(msg=f"Failed to find issue {issue}")
        module.fail_json(msg=f"Failed to send request to {url}: {info['msg']}")

    gh_obj = json.loads(response.read())

    if action == "get_status" or action is None:
        if module.check_mode:
            result.update(changed=True)
        else:
            result.update(changed=True, issue_status=gh_obj["state"])

    module.exit_json(**result)


if __name__ == "__main__":
    main()



# Copyright (c) Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import typing as t


def merge_settings_without_absent_nulls(
    existing_settings: dict[str, t.Any], desired_settings: dict[str, t.Any]
) -> dict[str, t.Any]:
    """
    Merges existing and desired settings into a new dictionary while excluding null values in desired settings that are absent in the existing settings.
    This ensures idempotency by treating absent keys in existing settings and null values in desired settings as equivalent, preventing unnecessary updates.

    Args:
      existing_settings (dict): Dictionary representing the current settings in Keycloak
      desired_settings (dict): Dictionary representing the desired settings

    Returns:
      dict: A new dictionary containing all entries from existing_settings and desired_settings,
      excluding null values in desired_settings whose corresponding keys are not present in existing_settings
    """

    existing = existing_settings or {}
    desired = desired_settings or {}

    return {**existing, **{k: v for k, v in desired.items() if v is not None or k in existing}}



# Copyright (c) 2021, Abhijeet Kasurde <akasurde@redhat.com>
# Copyright (c) 2018, Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
name: random_string
author:
  - Abhijeet Kasurde (@Akasurde)
short_description: Generates random string
version_added: '3.2.0'
description:
  - Generates random string based upon the given constraints.
  - Uses L(secrets.SystemRandom,https://docs.python.org/3/library/secrets.html#secrets.SystemRandom), so should be strong enough
    for cryptographic purposes.
options:
  length:
    description: The length of the string.
    default: 8
    type: int
  upper:
    description:
      - Possibly include uppercase letters in the string.
      - To ensure atleast one uppercase letter, set O(min_upper) to V(1).
    default: true
    type: bool
  lower:
    description:
      - Possibly include lowercase letters in the string.
      - To ensure atleast one lowercase letter, set O(min_lower) to V(1).
    default: true
    type: bool
  numbers:
    description:
      - Possibly include numbers in the string.
      - To ensure atleast one numeric character, set O(min_numeric) to V(1).
    default: true
    type: bool
  special:
    description:
      - Possibly include special characters in the string.
      - Special characters are taken from Python standard library C(string). See L(the documentation of
        string.punctuation,https://docs.python.org/3/library/string.html#string.punctuation)
        for which characters are used.
      - The choice of special characters can be changed to setting O(override_special).
      - To ensure atleast one special character, set O(min_special) to V(1).
    default: true
    type: bool
  min_numeric:
    description:
      - Minimum number of numeric characters in the string.
      - If set, overrides O(numbers=false).
    default: 0
    type: int
  min_upper:
    description:
      - Minimum number of uppercase alphabets in the string.
      - If set, overrides O(upper=false).
    default: 0
    type: int
  min_lower:
    description:
      - Minimum number of lowercase alphabets in the string.
      - If set, overrides O(lower=false).
    default: 0
    type: int
  min_special:
    description:
      - Minimum number of special character in the string.
    default: 0
    type: int
  override_special:
    description:
      - Override a list of special characters to use in the string.
      - If set O(min_special) should be set to a non-default value.
    type: str
  override_all:
    description:
      - Override all values of O(numbers), O(upper), O(lower), and O(special) with the given list of characters.
    type: str
  ignore_similar_chars:
    description:
      - Ignore similar characters, such as V(l) and V(1), or V(O) and V(0).
      - These characters can be configured in O(similar_chars).
    default: false
    type: bool
    version_added: 7.5.0
  similar_chars:
    description:
      - Override a list of characters not to be use in the string.
    default: "il1LoO0"
    type: str
    version_added: 7.5.0
  base64:
    description:
      - Returns base64 encoded string.
    type: bool
    default: false
  seed:
    description:
      - Seed for random string generator.
      - B(Note) that this drastically reduces the security of this plugin. First, when O(seed) is provided, a non-cryptographic random number generator is used.
        Second, if the seed does not contain enough entropy, the generated string is weak.
        B(Do not use the generated string as a password or a secure token when using this option!)
    type: str
    version_added: 11.3.0
"""

EXAMPLES = r"""
- name: Generate random string
  ansible.builtin.debug:
    var: lookup('community.general.random_string')
  # Example result: 'DeadBeeF'

- name: Generate random string with seed
  ansible.builtin.debug:
    var: lookup('community.general.random_string', seed=12345)
  # Example result: '6[~(2q5O'
  # NOTE: Do **not** use this string as a password or a secure token,
  #       unless you know exactly what you are doing!
  #       Specifying seed uses a non-secure random number generator.

- name: Generate random string with length 12
  ansible.builtin.debug:
    var: lookup('community.general.random_string', length=12)
  # Example result: 'Uan0hUiX5kVG'

- name: Generate base64 encoded random string
  ansible.builtin.debug:
    var: lookup('community.general.random_string', base64=True)
  # Example result: 'NHZ6eWN5Qk0='

- name: Generate a random string with 1 lower, 1 upper, 1 number and 1 special char (at least)
  ansible.builtin.debug:
    var: lookup('community.general.random_string', min_lower=1, min_upper=1, min_special=1, min_numeric=1)
  # Example result: '&Qw2|E[-'

- name: Generate a random string with all lower case characters
  ansible.builtin.debug:
    var: query('community.general.random_string', upper=false, numbers=false, special=false)
  # Example result: ['exolxzyz']

- name: Generate random hexadecimal string
  ansible.builtin.debug:
    var: query('community.general.random_string', upper=false, lower=false, override_special=hex_chars, numbers=false)
  vars:
    hex_chars: '0123456789ABCDEF'
  # Example result: ['D2A40737']

- name: Generate random hexadecimal string with override_all
  ansible.builtin.debug:
    var: query('community.general.random_string', override_all=hex_chars)
  vars:
    hex_chars: '0123456789ABCDEF'
  # Example result: ['D2A40737']
"""

RETURN = r"""
_raw:
  description: A one-element list containing a random string.
  type: list
  elements: str
"""

import base64
import random
import secrets
import string

from ansible.errors import AnsibleLookupError
from ansible.module_utils.common.text.converters import to_bytes, to_text
from ansible.plugins.lookup import LookupBase

from ansible_collections.community.general.plugins.plugin_utils._lookup import check_for_no_terms


class LookupModule(LookupBase):
    @staticmethod
    def get_random(random_generator, chars, length):
        if not chars:
            raise AnsibleLookupError("Available characters cannot be None, please change constraints")
        return "".join(random_generator.choice(chars) for dummy in range(length))

    @staticmethod
    def b64encode(string_value, encoding="utf-8"):
        return to_text(base64.b64encode(to_bytes(string_value, encoding=encoding, errors="surrogate_or_strict")))

    def run(self, terms, variables=None, **kwargs):
        number_chars = string.digits
        lower_chars = string.ascii_lowercase
        upper_chars = string.ascii_uppercase
        special_chars = string.punctuation

        self.set_options(var_options=variables, direct=kwargs)
        check_for_no_terms(self, terms=terms, direct=kwargs)

        length = self.get_option("length")
        base64_flag = self.get_option("base64")
        override_all = self.get_option("override_all")
        ignore_similar_chars = self.get_option("ignore_similar_chars")
        similar_chars = self.get_option("similar_chars")
        seed = self.get_option("seed")

        if seed is None:
            random_generator = secrets.SystemRandom()
        else:
            random_generator = random.Random(seed)

        values = ""
        available_chars_set = ""

        if ignore_similar_chars:
            number_chars = "".join([sc for sc in number_chars if sc not in similar_chars])
            lower_chars = "".join([sc for sc in lower_chars if sc not in similar_chars])
            upper_chars = "".join([sc for sc in upper_chars if sc not in similar_chars])
            special_chars = "".join([sc for sc in special_chars if sc not in similar_chars])

        if override_all:
            # Override all the values
            available_chars_set = override_all
        else:
            upper = self.get_option("upper")
            lower = self.get_option("lower")
            numbers = self.get_option("numbers")
            special = self.get_option("special")
            override_special = self.get_option("override_special")

            if override_special:
                special_chars = override_special

            if upper:
                available_chars_set += upper_chars
            if lower:
                available_chars_set += lower_chars
            if numbers:
                available_chars_set += number_chars
            if special:
                available_chars_set += special_chars

            mapping = {
                "min_numeric": number_chars,
                "min_lower": lower_chars,
                "min_upper": upper_chars,
                "min_special": special_chars,
            }

            for m in mapping:
                if self.get_option(m):
                    values += self.get_random(random_generator, mapping[m], self.get_option(m))

        remaining_pass_len = length - len(values)
        values += self.get_random(random_generator, available_chars_set, remaining_pass_len)

        shuffled_values = list(values)
        if seed is None:
            # Get pseudo randomization
            # Randomize the order
            random.shuffle(shuffled_values)

        if base64_flag:
            return [self.b64encode("".join(shuffled_values))]

        return ["".join(shuffled_values)]



#!/usr/bin/python
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

DOCUMENTATION = '''
---
module: azure_rm_ml_job
version_added: "3.17.0"
short_description: Create, Cancel, Archive or Restore an Azure Machine Learning Job
description:
    - Create, Cancel, Archive or Restore an Azure Machine Learning Job.
options:
    name:
        description:
            - Name of the Azure ML job.
        type: str
        required: true
    resource_group:
        description:
            - Name of resource group.
        required: true
        type: str
    ml_workspace:
        description:
            - Name of the Azure ML workspace.
        required: true
        type: str
    resource_definition:
        description:
            - Job definition in YAML
        type: str
        required: false
    description:
        description:
            - Description of the job.
        type: str
    state:
        description:
            - State of the Job. Use C(present) to create
              or update. C(cancel) will cancel.
              C(archive) to archive, C(restore) to restore.
        default: present
        type: str
        choices:
            - present
            - cancel
            - archive
            - restore

extends_documentation_fragment:
    - azure.azcollection.azure
    - azure.azcollection.azure_tags

author:
    - Bill Peck (@p3ck)
'''

EXAMPLES = '''
- name: Create ML Job
  azure.azcollection.azure_rm_ml_job:
    name: jobxxxxxxxxxx
    resource_group: xxxxx-ml-job
    ml_workspace: workspace-xxxxxxxxxx
    resource_definition: |
          $schema: https://azuremlschemas.azureedge.net/latest/commandJob.schema.json
          type: command
          description: Trains a simple neural network on the Fashion-MNIST dataset.
          experiment_name: "aml_command_cli"
          compute: azureml:cluster-cpu
          inputs:
            fashion_mnist:
              path: azureml:data-fashion-mnist@latest
          outputs:
            model:
              type: mlflow_model
          code: ./targets/azure_rm_ml_job/files/src
          environment:
            image: mcr.microsoft.com/azureml/openmpi5.0-ubuntu24.04:latest
            conda_file: ./targets/azure_rm_ml_job/files/conda.yml
          command: python train.py --data_dir ${{ '{{' }}inputs.fashion_mnist{{ '}}' }} --model_dir ${{ '{{' }}outputs.model{{ '}}' }}
'''

RETURN = '''
changed:
    description:
        - Whether the resource is changed.
    returned: always
    type: bool
    sample: false
ml_job:
    description:
        - Job that was just created or updated.
    returned: always
    type: dict
    sample: {
        "code": "azureml:/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/xxxxx-ml-job/providers/Microsoft.MachineLearningServices/workspaces/workspace-xxxxxxxxxx/codes/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/versions/1",
        "command": "python train.py --data_dir ${{inputs.fashion_mnist}} --model_dir ${{outputs.model}}",
        "compute": "azureml:cluster-cpu",
        "creation_context": {
            "created_at": "2026-03-31T17:55:36.317328+00:00",
            "created_by": "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
            "created_by_type": "Application"
        },
        "description": "Trains a simple neural network on the Fashion-MNIST dataset.",
        "display_name": "jobxxxxxxxxxx",
        "environment": "azureml:CliV2AnonymousEnvironment:xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
        "environment_variables": {},
        "experiment_name": "aml_command_cli",
        "id": "azureml:/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/xxxxx-ml-job/providers/Microsoft.MachineLearningServices/workspaces/workspace-xxxxxxxxxx/jobs/jobxxxxxxxxxx",
        "inputs": {
            "fashion_mnist": {
                "mode": "ro_mount",
                "path": "azureml:data-fashion-mnist:1",
                "type": "uri_folder"
            }
        },
        "name": "jobxxxxxxxxxx",
        "outputs": {
            "default": {
                "mode": "rw_mount",
                "path": "azureml://datastores/workspaceartifactstore/ExperimentRun/dcid.jobxxxxxxxxxx",
                "type": "uri_folder"
            },
            "model": {
                "mode": "rw_mount",
                "type": "mlflow_model"
            }
        },
        "parameters": {},
        "properties": {
            "ContentSnapshotId": "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
            "EndTimeUtc": "2026-03-31 18:18:41",
            "ProcessInfoFile": "azureml-logs/process_info.json",
            "ProcessStatusFile": "azureml-logs/process_status.json",
            "StartTimeUtc": "2026-03-31 18:14:56",
            "_azureml.ClusterName": "cluster-cpu",
            "_azureml.ComputeTargetType": "amlctrain"
        },
        "queue_settings": {
            "job_tier": "null"
        },
        "resources": {
            "instance_count": 1,
            "properties": {},
            "shm_size": "2g"
        },
        "services": {
            "Studio": {
                "endpoint": "https://ml.azure.com/runs/jobxxxxxxxxxx?wsid=/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourcegroups/xxxxx-ml-job/workspaces/workspace-xxxxxxxxxx&tid=xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
                "type": "Studio"
            },
            "Tracking": {
                "endpoint": "azureml://eastus.api.azureml.ms/mlflow/v1.0/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/xxxxx-ml-job/providers/Microsoft.MachineLearningServices/workspaces/workspace-xxxxxxxxxx?",
                "type": "Tracking"
            }
        },
        "status": "Completed",
        "tags": {},
        "type": "command"
    }
'''  # NOQA


try:
    import io
    import time
    from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common_ml import MLClientCommon
    from azure.core.exceptions import ResourceNotFoundError
    from azure.ai.ml.entities._load_functions import load_job
except ImportError:
    # This is handled in azure_rm_common
    pass


class AzureRMMLJob(MLClientCommon):

    UPDATE_OVERRIDES = []

    PARAM_OVERRIDES = UPDATE_OVERRIDES + ['name',
                                          'description',
                                          'tags']

    def __init__(self):

        self.module_arg_spec = dict(
            name=dict(
                type='str',
                required=True,
            ),
            resource_definition=dict(
                type='str',
                required=False,
            ),
            resource_group=dict(
                type='str',
                required=True,
            ),
            ml_workspace=dict(
                type='str',
                required=True,
            ),
            description=dict(
                type='str',
                required=False,
            ),
            state=dict(
                type='str',
                default='present',
                choices=['present', 'cancel', 'archive', 'restore'],
            ),
        )

        self._client = None
        self.ml_registry = None
        self.state = None

        self.results = dict(
            changed=False,
            compare=[],
            ml_job={}
        )

        required_if = [
            ('state', 'present', ['resource_definition'])
        ]

        super(AzureRMMLJob, self).__init__(self.module_arg_spec,
                                           supports_check_mode=True,
                                           required_if=required_if,
                                           )

    def exec_module(self, **kwargs):

        for key in list(self.module_arg_spec.keys()) + ['tags']:
            setattr(self, key, kwargs[key])

        changed = False
        ml_job_info = self.get(name=self.name)

        if self.state == 'present':
            if ml_job_info:
                # Update
                # No Updates are supported at this time.
                changed = False
                ml_job_info = self.entity_to_dict(ml_job_info)
            else:
                # Create
                params_override = self.update_params(kwargs, self.PARAM_OVERRIDES)

                # Generate ml_job based on ansible args passed in.
                resource_definition = io.StringIO(self.resource_definition)
                ml_job = load_job(source=resource_definition,
                                  params_override=params_override)

                changed = True
                if not self.check_mode:
                    response = self.client.jobs.create_or_update(job=ml_job)
                    ml_job_info = self.entity_to_dict(response)
        elif self.state == 'cancel':
            # Cancel
            if ml_job_info:
                changed = True
                if not self.check_mode:
                    self.cancel(name=self.name)
        elif self.state == 'archive':
            # Archive
            list_view_type = self.get_list_view_type('active')
            results = self.client.jobs.list(list_view_type=list_view_type)
            ml_jobs = [x.name for x in results]

            if self.name in ml_jobs:
                changed = True
                if not self.check_mode:
                    self.archive(name=self.name)
        elif self.state == 'restore':
            # Restore
            list_view_type = self.get_list_view_type('archived')
            results = self.client.jobs.list(list_view_type=list_view_type)
            ml_jobs = [x.name for x in results]

            if self.name in ml_jobs:
                changed = True
                if not self.check_mode:
                    self.restore(name=self.name)
                    # If we query too quick it hasn't fully restored.
                    # Seems the api shouldn't return until it's finished,
                    # but here we are.
                    time.sleep(1)
            ml_job_info = self.get(name=self.name,
                                   as_dict=True)

        self.results['ml_job'] = ml_job_info
        self.results['changed'] = changed
        return self.results

    def update_params(self, kwargs, param_overrides, ml_job=None):
        params_override = []

        # If ml_job is defined use those values as a base.
        # Update can't change some of these values but we need
        # to keep the same values as before the update.
        if ml_job:
            items = self.entity_to_dict(ml_job).items()
            params_override = [{k: v} for k, v in items]

        # Update values based on what param_overrides says we can update.
        for key in list(self.module_arg_spec.keys()):
            if kwargs[key] is not None and key in param_overrides:
                params_override.append({key: kwargs[key]})
        return params_override

    def get(self, name, as_dict=False):
        try:
            # Attempt to retrieve ml_job based on name
            ml_job_info = self.client.jobs.get(name=name)
            if as_dict:
                ml_job_info = self.entity_to_dict(ml_job_info)
        except ResourceNotFoundError:
            ml_job_info = None
        return ml_job_info

    def archive(self, name=None):
        try:
            self.client.jobs.archive(name=name)
        except Exception as e:  # pylint: disable=broad-exception-caught
            self.fail("Error archive the job instance: {0}".format(str(e)))

    def restore(self, name=None):
        try:
            self.client.jobs.restore(name=name)
        except Exception as e:  # pylint: disable=broad-exception-caught
            self.fail("Error restore the job instance: {0}".format(str(e)))

    def cancel(self, name=None):
        try:
            response = self.client.jobs.begin_cancel(name)
            ml_job_object = self.get_poller_result(response)
            return self.entity_to_dict(ml_job_object)
        except Exception as e:  # pylint: disable=broad-exception-caught
            self.fail("Error cancelling  the job instance: {0}".format(str(e)))


def main():
    AzureRMMLJob()


if __name__ == '__main__':
    main()
